#
# Revelation - a password manager for GNOME 2
# http://oss.codepoet.no/revelation/
# $Id$
#
# Module for handling Figaro's Password Manager data
#
#
# Copyright (c) 2003-2006 Erik Grinaker
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from . import base
from revelation import data, entry, util

import math, random, string, xml.dom.minidom

from xml.parsers.expat import ExpatError
from Cryptodome.Cipher import Blowfish
from Cryptodome.Hash import MD5
import Cryptodome.Random as Random


class FPM(base.DataHandler):
    "Data handler for Figaro's Password Manager data"

    name        = "Figaro's Password Manager"
    importer    = True
    exporter    = True
    encryption  = True


    def __init__(self):
        base.DataHandler.__init__(self)


    def __decrypt(self, cipher, data):
        "Decrypts data"

        if isinstance(data, str):
            data = data.encode()

        # decode ascii armoring
        decoded = b""

        for i in range(len(data) // 2):
            high = data[2 * i] - ord("a")
            low =  data[2 * i + 1] - ord("a")
            decoded += bytes((high * 16 + low,))
        data = decoded

        # decrypt data
        data = cipher.decrypt(data)

        # unrotate field
        blocks = int(math.ceil(len(data) / float(8)))
        plain = b""

        for offset in range(8):
            for block in range(blocks):
                plain += bytes((data[block * 8 + offset],))

        return plain.split(b"\x00")[0]


    def __encrypt(self, cipher, data):
        "Encrypts data"

        # get data sizes
        blocks = (len(data) // 7) + 1
        size = 8 * blocks

        # add noise
        rand = Random.new()
        data += b'\x00' + rand.read(size - len(data) - 1)

        # rotate data
        rotated = b""
        for block in range(blocks):
            for offset in range(8):
                rotated += bytes((data[offset * blocks + block],))

        data = rotated

        # encrypt data
        data = cipher.encrypt(data)

        # ascii-armor data
        res = b""

        for i in range(len(data)):
            high = data[i] // 16
            low = data[i] - high * 16
            res += bytes((ord("a") + high, ord("a") + low))

        data = res


        return data


    def check(self, input):
        "Checks if the data is valid"

        try:
            if input is None:
                raise base.FormatError

            dom = xml.dom.minidom.parseString(input.strip())

            if dom.documentElement.nodeName != "FPM":
                raise base.FormatError

            minversion = dom.documentElement.attributes["min_version"].nodeValue

            if int(minversion.split(".")[1]) > 58:
                raise base.VersionError


        except ExpatError:
            raise base.FormatError

        except ( KeyError, IndexError ):
            raise base.FormatError


    def detect(self, input):
        "Checks if this handler can handle the given data"

        try:
            self.check(input)
            return True

        except ( base.FormatError, base.VersionError, base.DataError ):
            return False


    def export_data(self, entrystore, password):
        "Exports data from an entrystore"

        # set up encryption engine
        salt = bytes( [ random.choice(string.ascii_lowercase.encode()) for i in range(8) ] )
        password = MD5.new(salt + password.encode()).digest()

        cipher = Blowfish.new(password, Blowfish.MODE_ECB)


        # generate data
        xml = "<?xml version=\"1.0\" ?>\n"
        xml += "<FPM full_version=\"00.58.00\" min_version=\"00.58.00\" display_version=\"00.58.00\">\n"
        xml += "    <KeyInfo salt=\"%s\" vstring=\"%s\" />\n" % ( salt.decode(), self.__encrypt(cipher, b"FIGARO").decode() )
        xml += "    <LauncherList></LauncherList>\n"
        xml += "    <PasswordList>\n"

        iter = entrystore.iter_children(None)

        while iter is not None:
            e = entrystore.get_entry(iter)

            if type(e) != entry.FolderEntry:
                e = e.convert_generic()

                xml += "        <PasswordItem>\n"
                xml += "            <title>%s</title>\n" % self.__encrypt(cipher, e.name.encode()).decode()
                xml += "            <url>%s</url>\n" % self.__encrypt(cipher, e.get_field(entry.HostnameField).value.encode()).decode()
                xml += "            <user>%s</user>\n" % self.__encrypt(cipher, e.get_field(entry.UsernameField).value.encode()).decode()
                xml += "            <password>%s</password>\n" % self.__encrypt(cipher, e.get_field(entry.PasswordField).value.encode()).decode()
                xml += "            <notes>%s</notes>\n" % self.__encrypt(cipher, e.description.encode()).decode()

                path = entrystore.get_path(iter).to_string()

                if len(path) > 1:
                    foldername = entrystore.get_entry(entrystore.get_iter(path[0])).name
                    xml += "            <category>%s</category>\n" % self.__encrypt(cipher, foldername.encode()).decode()

                else:
                    xml += "            <category></category>\n"

                xml += "            <launcher></launcher>\n"
                xml += "        </PasswordItem>\n"

            iter = entrystore.iter_traverse_next(iter)


        xml += "    </PasswordList>\n"
        xml += "</FPM>\n"


        return xml


    def import_data(self, input, password):
        "Imports data into an entrystore"

        try:

            # check and load data
            self.check(input)
            dom = xml.dom.minidom.parseString(input.strip())

            if dom.documentElement.nodeName != "FPM":
                raise base.FormatError


            # set up decryption engine, and check if password is correct
            keynode = dom.documentElement.getElementsByTagName("KeyInfo")[0]
            salt = keynode.attributes["salt"].nodeValue.encode()
            vstring = keynode.attributes["vstring"].nodeValue.encode()

            password = MD5.new(salt + password.encode()).digest()
            cipher = Blowfish.new(password, Blowfish.MODE_ECB)

            if self.__decrypt(cipher, vstring) != b"FIGARO":
                raise base.PasswordError

        except ExpatError:
            raise base.FormatError


        except ( IndexError, KeyError ):
            raise base.FormatError


        # import entries into entrystore
        entrystore = data.EntryStore()
        folders = {}

        for node in dom.getElementsByTagName("PasswordItem"):

            parent = None
            e = entry.GenericEntry()

            for fieldnode in [ node for node in node.childNodes if node.nodeType == node.ELEMENT_NODE ]:

                content = self.__decrypt(cipher, util.dom_text(fieldnode)).decode()

                if content == "":
                    continue

                elif fieldnode.nodeName == "title":
                    e.name = content

                elif fieldnode.nodeName == "user":
                    e.get_field(entry.UsernameField).value = content

                elif fieldnode.nodeName == "url":
                    e.get_field(entry.HostnameField).value = content

                elif fieldnode.nodeName == "password":
                    e.get_field(entry.PasswordField).value = content

                elif fieldnode.nodeName == "notes":
                    e.description = content

                elif fieldnode.nodeName == "category":

                    if content in folders:
                        parent = folders[content]

                    else:
                        folderentry = entry.FolderEntry()
                        folderentry.name = content

                        parent = entrystore.add_entry(folderentry)
                        folders[content] = parent

            entrystore.add_entry(e, parent)

        return entrystore






#
# Revelation - a password manager for GNOME 2
# http://oss.codepoet.no/revelation/
# $Id$
#
# Module for handling plain text files
#
#
# Copyright (c) 2003-2006 Erik Grinaker
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from . import base
from revelation import entry

import time


class PlainText(base.DataHandler):
    "Data handler for plain text files"

    name        = "Plain text"
    importer    = False
    exporter    = True
    encryption  = False


    def export_data(self, entrystore, password = None):
        "Exports data to a plain text file"

        text = ""
        # fetch and sort entries
        iter = entrystore.iter_nth_child(None, 0)

        while iter is not None:
            e = entrystore.get_entry(iter)

            if type(e) == entry.FolderEntry:
                text += "%s [Folder]\n" % e.name
                text += "%s\n" % e.description
                text += "%s\n" % time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(e.updated))
            else:
                text += "%s [%s]\n" % ( e.name, e.typename )
                text += e.description != "" and "%s\n" % e.description or ""
                text += "%s\n" % time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(e.updated))

                fields = [ field for field in e.fields if field.value != "" ]

                if len(fields) > 0:
                    text += "\n"
                    maxlen = max([ len(field.name) for field in fields ])

                    for field in fields:
                        text += "- " + field.name + ": " + (" " * (maxlen - len(field.name))) + field.value + "\n"

            text += "\n\n"

            iter = entrystore.iter_traverse_next(iter)

        return text




#
# Revelation
# Module for importing data from KeepassXC CSV export file
#
# Copyright (c) 2020 Mikel Olasagasti Uranga <mikel@olasagasti.info>
# Copyright (c) 2006 Devan Goodwin <dgoodwin@dangerouslyinc.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import csv
import time

from revelation import data, entry
from . import base


class KeepassXCCSV(base.DataHandler):
    """
    Data handler for CSV files generated by KeepassXC.

    CSV format is defined in KeepassXC's CsvExporter.cpp

    https://github.com/keepassxreboot/keepassxc/blob/develop/src/format/CsvExporter.cpp

    CSV field description:

    Type:       Ignored
    Field 1:    Group
    Field 2:    Title
    Field 3:    Username
    Field 4:    Password
    Field 5:    URL
    Field 6:    Notes
    # since KeepassXC-2.6.2
    Field 7:    TOTP
    Field 8:    Icon
    Field 9:    Last Modified
    Field 10:   Created
    """

    name = "KeepassXC CSV"
    importer = True
    exporter = False
    encryption = False

    def import_data(self, input, password):
        " Import data from a file into the entry store"

        entrystore = data.EntryStore()

        # Maintain a hash of folder names to folder entries so we
        # can use each category encountered to create a new folder
        # by that name, or use an existing one if we've already
        # created it:
        folders = {}

        for line in input.splitlines()[1:]:
            f_csv = csv.reader([line.decode()])
            for row in f_csv:

                # Raise FormatError if we don't have all 9 fields
                # KeepasXC 2.5 to 2.6.1 has 6 fields
                # KeepasXC 2.6.2 has 10 fields
                if len(row) != 6 and len(row) != 10:
                    raise base.FormatError

                # If URL is present create WebEntry
                if row[4]:
                    new_entry = entry.WebEntry()
                    new_entry[entry.URLField] = row[4]
                else:
                    new_entry = entry.GenericEntry()

                new_entry.name = row[1]
                new_entry[entry.UsernameField] = row[2]
                new_entry[entry.PasswordField] = row[3]
                new_entry.notes = row[5]

                # TODO As Last modified and creationtime are from newer
                # keepassXC releases, set to current time for now
                new_entry.updated = time.time()

                # Create and/or add to folder
                # TODO split folder name and correctly group them
                if row[0] in folders:
                    parent = folders[row[0]]

                else:
                    folder = entry.FolderEntry()
                    folder.name = row[0]
                    parent = entrystore.add_entry(folder)
                    folders[row[0]] = parent

                # Add the entry
                entrystore.add_entry(new_entry, parent)

        return entrystore



#
# Revelation - a password manager for GNOME 2
# http://oss.codepoet.no/revelation/
# $Id$
#
# Module for basic datahandler functionality
#
#
# Copyright (c) 2003-2006 Erik Grinaker
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

class Error(Exception):
    "Base exception for data handler"
    pass

class DataError(Error):
    "Exception for invalid data"
    pass

class FormatError(Error):
    "Exception for invalid file formats"
    pass

class PasswordError(Error):
    "Exception for wrong password"
    pass

class VersionError(Error):
    "Exception for unknown versions"
    pass



class DataHandler(object):
    "A datahandler base class, real datahandlers are subclassed from this"

    name        = None
    importer    = False
    exporter    = False
    encryption  = False


    def __init__(self):
        pass


    def check(self, input):
        "Fallback method, subclasses should override this"

        pass


    def detect(self, input):
        "Fallback method, subclasses should override this"

        return False


    def export_data(self, entrystore, password):
        "Fallback method, subclasses should override this"

        return ""


    def import_data(self, input, password):
        "Fallback method, subclasses should override this"

        pass




#
# Revelation - a password manager for GNOME 2
# http://oss.codepoet.no/revelation/
# $Id$
#
# Module with various utility functions
#
#
# Copyright (c) 2003-2006 Erik Grinaker
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import datetime, pwquality, gettext, math, os, random, shlex, string, traceback
from io import StringIO

_ = gettext.gettext


class SubstFormatError(Exception):
    "Exception for parse_subst format errors"
    pass


class SubstValueError(Exception):
    "Exception for missing values in parse_subst"
    pass



def check_password(password):
    "Checks if a password is valid"

    # check for length
    if len(password) < 6:
        raise ValueError(_('is too short'))


    # check for entropy
    pwlen   = len(password)
    ent = entropy(password)
    idealent = entropy_ideal(pwlen)

    if (pwlen < 100 and ent / idealent < 0.8) or (pwlen >= 100 and ent < 5.3):
        raise ValueError(_('isn\'t varied enough'))


    # check the password strength
    lc, uc, d, o = 0, 0, 0, 0

    for c in password:
        if c in string.ascii_lowercase:
            lc += 1

        elif c in string.ascii_uppercase:
            uc += 1

        elif c in string.digits:
            d += 1

        else:
            o += 1

    classcount = [ lc, uc, d, o ]
    classcount.sort()
    classcount.reverse()

    cred = sum([ count * (1 + (weight ** 2.161 / 10)) for weight, count in zip(range(1, len(classcount) + 1), classcount) ])

    if cred < 10:
        raise ValueError(_('is too weak'))


    # check if the password is a palindrome
    for i in range(len(password)):
        if password[i] != password[-i - 1]:
            break

    else:
        raise ValueError(_('is a palindrome'))


    # check password with pwquality
    try:
        if len(password) < 100:
            pwquality.PWQSettings().check(password)

    except (ValueError, pwquality.PWQError)  as reason:

        # modify reason
        if isinstance(reason, pwquality.PWQError):
            reason = reason.args[1].replace("The password ", "")
        reason = str(reason).strip()
        reason = reason.replace("simplistic/systematic", "systematic")
        reason = reason.replace(" dictionary", "")

        if reason[:3] == "it ":
            reason = reason[3:]

        if reason[:5] == "it's ":
            reason = "is " + reason[5:]

        raise ValueError(reason)

    except IOError:
        pass


def dom_text(node):
    "Returns text content of a DOM node"

    text = ""

    for child in node.childNodes:
        if child.nodeType == node.TEXT_NODE:
            text += child.nodeValue

    return text


def entropy(string):
    "Calculates the Shannon entropy of a string"

    # get probability of chars in string
    prob = [ float(string.count(c)) / len(string) for c in dict.fromkeys(list(string)) ]

    # calculate the entropy
    entropy = - sum([ p * math.log(p) / math.log(2.0) for p in prob ])

    return entropy


def entropy_ideal(length):
    "Calculates the ideal Shannon entropy of a string with given length"

    prob = 1.0 / length

    return -1.0 * length * prob * math.log(prob) / math.log(2.0)


def escape_markup(string):
    "Escapes a string so it can be placed in a markup string"

    if string is None:
        return ""

    string = string.replace("&", "&amp;")
    string = string.replace("<", "&lt;")
    string = string.replace(">", "&gt;")

    return string


def execute(command):
    "Runs a command, returns its status code and output"

    p = os.popen(command, "r")
    output = p.read()
    status = p.close()

    if status is None:
        status = 0

    status = status >> 8

    return output, status


def execute_child(command):
    "Runs a command as a child, returns its process ID"

    items = shlex.split(command)

    return os.spawnvp(os.P_NOWAIT, items[0], items)


def generate_password(length, punctuation):
    "Generates a password"

    # set up character sets
    d   = string.digits.translate(str.maketrans("", "", "015"))
    lc  = string.ascii_lowercase.translate(str.maketrans("", "", "lqg"))
    uc  = string.ascii_uppercase.translate(str.maketrans("", "", "IOS"))
    fullset = d + uc + lc

    if punctuation:
        p   = string.punctuation
        fullset = fullset + p
        charsets = (
            ( d,    0.15 ),
            ( uc,   0.24 ),
            ( lc,   0.24 ),
            ( p,    0.15 ),
        )
    else:
        charsets = (
            ( d,    0.15 ),
            ( uc,   0.24 ),
            ( lc,   0.24 ),
        )


    # function for generating password
    def genpw(length):
        password = []

        for set, share in charsets:
            password.extend([ random.choice(set) for i in range(int(round(length * share))) ])

        while len(password) < length:
            password.append(random.choice(fullset))

        random.shuffle(password)

        return "".join(password)


    # check password, and regenerate if needed
    while True:
        try:
            password = genpw(length)

            if length <= 6:
                return password

            check_password(password)

            return password

        except ValueError:
            continue


def pad_right(string, length, padchar = " "):
    "Right-pads a string to a given length"

    if string is None:
        return None

    if len(string) >= length:
        return string

    return string + ((length - len(string)) * padchar)


def parse_subst(string, map):
    "Parses a string for substitution variables"

    result = ""

    pos = 0
    while pos < len(string):

        char = string[pos]
        next = pos + 1 < len(string) and string[pos + 1] or ""


        # handle normal characters
        if char != "%":
            result += char
            pos += 1


        # handle % escapes (%%)
        elif next == "%":
            result += "%"
            pos += 2


        # handle optional substitution variables
        elif next == "?":
            if string[pos + 2] in map:
                result += map[string[pos + 2]]
                pos += 3

            else:
                raise SubstFormatError


        # handle optional substring expansions
        elif next == "(":

            try:
                result += parse_subst(string[pos + 2:string.index("%)", pos + 1)], map)

            except ValueError:
                raise SubstFormatError

            except SubstValueError:
                pass

            pos = string.index("%)", pos + 1) + 2


        # handle required ("normal") substitution variables
        elif next in map:

            if not map[next]:
                raise SubstValueError

            result += map[next]
            pos += 2


        # otherwise, it's a format error
        else:
            raise SubstFormatError


    return result


def random_string(length):
    "Generates a random string"

    s = ""
    for i in range(length):
        s += chr(int(random.random() * 255))

    return s


def time_period_rough(start, end):
    "Returns the rough period from start to end in human-readable format"

    if end < start:
        return _('%i seconds') % 0

    start   = datetime.datetime.utcfromtimestamp(float(start))
    end = datetime.datetime.utcfromtimestamp(float(end))
    delta   = end - start


    if delta.days >= 365:
        period  = delta.days / 365
        unit    = period != 1 and _('years') or _('year')

    elif delta.days >= 31 or (end.month != start.month and (end.day > start.day or (end.day == start.day and end.time() >= start.time()))):
        period = ((end.year - start.year) * 12) + end.month - start.month

        if end.day < start.day or (end.day == start.day and end.time() < start.time()):
            period -= 1

        unit = period != 1 and _('months') or _('month')

    elif delta.days >= 7:
        period  = delta.days / 7
        unit    = period != 1 and _('weeks') or _('week')

    elif delta.days >= 1:
        period  = delta.days
        unit    = period != 1 and _('days') or _('day')

    elif delta.seconds >= 3600:
        period  = delta.seconds / 3600
        unit    = period != 1 and _('hours') or _('hour')

    elif delta.seconds >= 60:
        period  = delta.seconds / 60
        unit    = period != 1 and _('minutes') or _('minute')

    else:
        period  = delta.seconds
        unit    = period != 1 and _('seconds') or _('second')


    return "%d %s" % ( period, unit )


def trace_exception(type, value, tb):
    "Returns an exception traceback as a string"

    trace = StringIO()
    traceback.print_exception(type, value, tb, None, trace)

    return trace.getvalue()


def unescape_markup(string):
    "Unescapes a string to get literal values"

    string = string.replace("&amp;", "&")
    string = string.replace("&lt;", "<")
    string = string.replace("&gt;", ">")

    return string




"""
Implements the LUKS (Linux Unified Key Setup) Version 1.0
on disk specification inside a file.

http://luks.endorphin.org/

LUKS offers:
    * compatiblity via standardization,
    * secure against low entropy attacks,
    * support for multiple keys,
    * effective passphrase revocation,

This module is compatible with dm-crypt and cryptsetup tools for the Linux
kernel, as long as hashSpec="sha1" is used. Loopback files or partitions created
with the linux kernel can be decrypted using this module.  FreeOTFE
(http://www.freeotfe.org/) should provide support for reading and writing on
Windows.

This module has one class LuksFile.

Loading a LUKS disk image (use both, one after another):
- load_from_file(file)
- open_any_key(password)

Creating a new LUKS disk image (use both):
- create(file, cipherName, cipherMode, hashSpec, masterSize, stripes)
- set_key(0, password, iterations)

Once a file is unlocked (either because it was just created or
open_any_key() returned True), you can perform the key operations:
- enabled_key_count()
- key_information(keyIndex)
- set_key(keyIndex, password, iterations)
- delete_key(keyIndex)

Once a file is unlocked, you can perform data encryption/decryption with
- data_length()
- encrypt_data(offset, data)
- decrypt_data(offset, length)
- truncate(length)

Finally, to close the file:
- close()

Copyright 2006 John Lenz <lenz@cs.wisc.edu>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

http://www.gnu.org/copyleft/gpl.html
"""

import os, math, struct, stat, hashlib

import Cryptodome.Random as Random
from Cryptodome.Cipher import AES
from . import PBKDFv2, AfSplitter

class LuksError(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)

class LuksFile:
    """Implements the LUKS (Linux Unified Key Setup) Version 1.0 http://luks.endorphin.org/"""

    LUKS_FORMAT = ">6sH32s32s32sII20s32sI40s"
    LUKS_MAGIC = b"LUKS\xba\xbe"

    LUKS_KEY_DISABLED = 0x0000DEAD
    LUKS_KEY_ENABLED = 0x00AC71F3

    SECTOR_SIZE = 512.0

    KEY_STRIPES = 4000
    SALT_SIZE = 32
    DIGEST_SIZE = 20

    def __init__(self):
        self.file = None
        self.masterKey = None
        self.ivGen = None

    # Read the header from the file descriptor
    def load_from_file(self, file):
        """Initialize this LuksFile class from the file.

        The file parameter should be an object implementing the File Object API
        This function will error if the file is not a LUKS partition (the LUKS_MAGIC does not match)
        """

        if self.file != None:
            raise LuksError("This LuksFile has already been initialized")

        # Read the main parameters
        self.file = file
        self.file.seek(0)

        self.magic, \
        self.version, \
        cipherName, \
        cipherMode, \
        hashSpec, \
        self.payloadOffset, \
        self.keyBytes, \
        self.mkDigest, \
        self.mkDigestSalt, \
        self.mkDigestIterations, \
        self.uuid = \
        struct.unpack(self.LUKS_FORMAT, self.file.read(208))

        cipherName = cipherName.decode()
        cipherMode = cipherMode.decode()
        hashSpec = hashSpec.decode()

        # check magic
        if self.magic != self.LUKS_MAGIC:
            self.file = None
            raise LuksError("%s is not a LUKS data file" % filename)

        # Check the hash and cipher
        self.hashSpec = hashSpec.strip(" \x00")
        self._check_cipher(cipherName.strip(" \x00"), cipherMode.strip(" \x00"))

        # Load the key information
        self.keys = [None] * 8
        for i in range(0, 8):
            self.keys[i] = self._key_block()
            self.keys[i].load_from_str(self.file.read(48))

        # set the digest to be the correct size
        self.mkDigest = self.mkDigest[:hashlib.new(self.hashSpec).digest_size]

        self.masterKey = None

    # Generate a new header
    def create(self, file, cipherName, cipherMode, hashSpec, masterSize, stripes):
        """Initializes the file class passed in with the LUKS header

        Parameters
           cipherName: aes, cast5, blowfish
           cipherMode: cbc-plain, cbc-essiv:<hash>
           hashSpec: sha1, sha256, md5, ripemd160
           masterSize: length of the master key in bytes (must match cipher)
           stripes: number of stripes when Af Splitting keys

        For compatibility with the Linux kernel dm-crypt, hashSpec must equal "sha1"

        cbc-plain uses the sector number as the IV.  This has a weakness: an attacker
        may be able to detect the existance of watermarked files.
        cbc-essiv:<hash> protects against the weakness in cbc-plain, but is
        slightly slower. The digest size of the hash function passed to cbc-essiv
        must match the key size of the cipher.
        aes-cbc-essiv:sha256 works, while aes-cbc-essiv:sha1 does not

        For more information about the details of the attacks and risk assesment, see
        http://clemens.endorphin.org/LinuxHDEncSettings
        """

        if self.file != None:
            raise LuksError("This LuksFile has already been initialized")

        self._check_cipher(cipherName, cipherMode)

        self.magic = self.LUKS_MAGIC
        self.version = 1
        self.mkDigestIterations = 10
        self.keyBytes = masterSize
        self.hashSpec = hashSpec

        rand = Random.new()

        # Generate the salt
        self.mkDigestSalt = rand.read(self.SALT_SIZE)

        # Generate a random master key
        self.masterKey = rand.read(self.keyBytes)
        self.ivGen.set_key(self.masterKey)

        # generate the master key digest
        pbkdf = PBKDFv2.PBKDFv2()
        self.mkDigest = pbkdf.makeKey(self.masterKey, self.mkDigestSalt, self.mkDigestIterations, hashlib.new(self.hashSpec).digest_size, self.hashSpec)

        # init the key information
        currentSector = math.ceil(592.0 / self.SECTOR_SIZE)
        alignSectors = 4096 / self.SECTOR_SIZE
        blocksPerStripe = math.ceil(float(self.keyBytes * stripes) / self.SECTOR_SIZE)

        self.keys = [None] * 8
        for i in range(0, 8):
            if currentSector % alignSectors > 0:
                currentSector += alignSectors - currentSector % alignSectors
            self.keys[i] = self._key_block()
            self.keys[i].create(currentSector, stripes, self.LUKS_KEY_DISABLED)
            currentSector += blocksPerStripe

        # Set the data offset
        if currentSector % alignSectors > 0:
            currentSector += alignSectors - currentSector % alignSectors
        self.payloadOffset = int(currentSector)

        # Generate a UUID for this file
        self._uuidgen(rand)

        # Create a new file, and save the header into it
        self.file = file
        self._save_header()

        # Write FF into all the key slots
        FFData = bytes([255] * int(self.SECTOR_SIZE))
        for i in range(0, 8):
            self.file.seek(int(self.keys[i].keyMaterialOffset))
            for sector in range(0, int(blocksPerStripe)):
                self.file.write(FFData)

        # Flush the file to disk
        try:
            self.file.flush()
            os.fsync(self.file.fileno())
        except:
            # We might get an error because self.file.fileno() does not exist on StringIO
            pass

    def set_key(self, keyIndex, password, iterations, checkMinStripes=0):
        """Sets the key block at keyIndex using password

        Sets the key block at keyIndex using password, hashed iterations
        times using PBKDFv2 (RFC2104).  This LuksFile must be unlocked by
        calling open_any_key() before calling this function.
        checkMinStripes is used to detect basic header manipulation, since
        the number of stripes for the key is set by create(), before
        we write a password to disk we make sure the key is not weak because
        of a small number of stripes.

        This function will raise an error if the key is already enabled.
        """

        # Some checks
        if self.masterKey == None:
            raise LuksError("A key has not been unlocked.  Call open_any_key() first.")

        if keyIndex < 0 or keyIndex > 7:
            raise LuksError("keyIndex out of range")

        key = self.keys[keyIndex]
        if key.active != self.LUKS_KEY_DISABLED:
            raise LuksError("Key is active.  Delete the key and try again")

        if checkMinStripes == 0: checkMinStripes = self.KEY_STRIPES
        if key.stripes < checkMinStripes:
            raise LuksError("Key section %i contains too few stripes.  Header manipulation?" % keyIndex)

        key.passwordIterations = iterations

        # Generate a random salt for this key
        rand = Random.new()
        key.passwordSalt = rand.read(self.SALT_SIZE)

        # Hash the key using PBKDFv2
        pbkdf = PBKDFv2.PBKDFv2()
        derived_key = pbkdf.makeKey(password.encode(), key.passwordSalt, key.passwordIterations, self.keyBytes, self.hashSpec)

        # Split the key into key.stripes
        AfKey = AfSplitter.AFSplit(self.masterKey, key.stripes, self.hashSpec)

        AfKeySize = len(AfKey)
        if AfKeySize != key.stripes * self.keyBytes:
            raise LuksError("ERROR: AFSplit did not return the correct length of key")

        # Set the key for IV generation
        self.ivGen.set_key(derived_key)

        # Encrypt the splitted key using the hashed password
        AfSectors = int(math.ceil(float(AfKeySize) / self.SECTOR_SIZE))
        for sector in range(0, AfSectors):
            self._encrypt_sector(derived_key, key.keyMaterialOffset + sector, sector, \
                   AfKey[int(sector*self.SECTOR_SIZE):int((sector+1)*self.SECTOR_SIZE)])

        key.active = self.LUKS_KEY_ENABLED

        # Reset the key used for to IV generation in data mode
        self.ivGen.set_key(self.masterKey)

        self._save_header()

    def open_key(self, keyIndex, password):
        """Open a specific keyIndex using password.  Returns True on success"""

        if self.file == None:
            raise LuksError("LuksFile has not been initialized")

        if keyIndex < 0 or keyIndex > 7:
            raise LuksError("keyIndex is out of range")

        key = self.keys[keyIndex]

        if key.active != self.LUKS_KEY_ENABLED:
            return False

        # Hash the password using PBKDFv2
        pbkdf = PBKDFv2.PBKDFv2()
        derived_key = pbkdf.makeKey(password.encode(), key.passwordSalt, key.passwordIterations, self.keyBytes, self.hashSpec)

        # Setup the IV generation to use this key
        self.ivGen.set_key(derived_key)

        # Decrypt the master key data using the hashed password
        AfKeySize = key.stripes * self.keyBytes
        AfSectors = int(math.ceil(float(AfKeySize) / self.SECTOR_SIZE))
        AfKey = b""
        for sector in range(0, AfSectors):
            AfKey += self._decrypt_sector(derived_key, key.keyMaterialOffset + sector, sector)
        AfKey = AfKey[0:AfKeySize]

        # Merge the decrypted master key
        masterKey = AfSplitter.AFMerge(AfKey, key.stripes, self.hashSpec)

        # Check if the password was the correct one, by checking the master key digest
        checkDigest = pbkdf.makeKey(masterKey, self.mkDigestSalt, self.mkDigestIterations, hashlib.new(self.hashSpec).digest_size, self.hashSpec)

        # Since the header only stores DIGEST_SIZE (which is smaller than sha256 digest size)
        #   trim the digest to DIGEST_SIZE
        checkDigest = checkDigest[:self.DIGEST_SIZE]

        if checkDigest != self.mkDigest:
            return False

        self.masterKey = masterKey
        self.ivGen.set_key(self.masterKey)
        return True


    def open_any_key(self, password):
        """Try to open any enabled key using the provided password.  Returns index number on success, or None"""

        if self.file == None:
            raise LuksError("LuksFile has not been initialized")

        for i in range(0, 8):
            if self.open_key(i, password):
                return i
        return None

    def enabled_key_count(self):
        """Returns the number of enabled key slots"""

        if self.file == None:
            raise LuksError("LuksFile has not been initialized")

        cnt = 0
        for i in range(0, 8):
            if self.keys[i].active == self.LUKS_KEY_ENABLED:
                cnt += 1
        return cnt

    def key_information(self, keyIndex):
        """Returns a tuple of information about the key at keyIndex (enabled, iterations, stripes)"""

        if self.file == None:
            raise LuksError("LuksFile has not been initialized")

        if keyIndex < 0 or keyIndex > 7:
            raise LuksError("keyIndex out of range")

        key = self.keys[keyIndex]
        active = (key.active == self.LUKS_KEY_ENABLED)
        return (active, key.passwordIterations, key.stripes)

    def delete_key(self, keyIndex):
        """Delete the key located in slot keyIndex.  WARNING! This is NOT a secure delete

        Warning! If keyIndex is the last enabled key, the data will become unrecoverable

        This function will not securely delete the key.  Because this class is designed
        for reading and writing to a file, there is no guarante that writing over the data
        inside the file will destroy it.  If you have a key leaked, you need to investigate
        other methods of securely destroying data, including destroying the entire file system
        and disk this file was located on, and any backups of this file that were created
        (depending on your needs).  The good news is, because of the way LUKS encrypts the
        master key, only one bit in the master key needs to be destoryed.  But the same bit
        needs to be destroyed in all copies, including (possibly) the journal, bad remapped
        blocks on the disk, etc.

        If you would like to continue using this encrypted file, you need to set_key() a new
        key, delete_key() the leaked key, and then copy the file into a new file on a
        different disk and filesystem.  This class writes "FF" to the key location during
        delete_key(), so during the copy the new disk will just get "FF" in the key location,
        which will be unrecoverable on the new disk.
        """

        if self.file == None:
            raise LuksError("LuksFile has not been initialized")

        if keyIndex < 0 or keyIndex > 7:
            raise LuksError("keyIndex out of range")

        key = self.keys[keyIndex]

        # Start and end offset of the key material
        startOffset = key.keyMaterialOffset
        endOffset = startOffset + int(math.ceil((self.keyBytes * key.stripes) / self.SECTOR_SIZE))

        # Just write "FF" into the locations
        for i in range(startOffset, endOffset):
            self.file.seek(int(i * self.SECTOR_SIZE))
            self.file.write("\xFF" * int(self.SECTOR_SIZE))

        try:
            self.file.flush()
            os.fsync(self.file.fileno())
        except:
            # We might get an error because self.file.fileno() does not exist on StringIO
            pass

        key.active = self.LUKS_KEY_DISABLED
        key.passwordIterations = 0
        key.passwordSalt = ''

        self._save_header()

    def close(self):
        """Close the underlying file descriptor, and discard the cached master key used for decryption"""

        if self.ivGen != None:
            self.ivGen.set_key(b"")
        if self.file != None:
            self.file.close()

        self.file = None
        self.masterKey = None
        self.ivGen = None

    def data_length(self):
        """Returns the total data length"""

        if self.file == None:
            raise LuksError("LuksFile has not been initialized")

        # Seek to the end of the file, and use tell()
        self.file.seek(0, 2)
        fLen = self.file.tell()
        return fLen - int(self.payloadOffset * self.SECTOR_SIZE)

    def truncate(self, length):
        """Truncate the file so that the data is maximum of length in size"""

        if self.file == None:
            raise LuksError("LuksFile has not been initialized")

        if length % self.SECTOR_SIZE != 0:
            raise LuksError("length must be a multiple of %s" % self.SECTOR_SIZE)

        if length < 0:
            raise LuksError("length must be positive")

        self.file.truncate(int(self.payloadOffset * self.SECTOR_SIZE) + length)

    def encrypt_data(self, offset, data):
        """Encrypt data into the file.

        Offset is a zero indexed location in the data to write.
        Both offset and len(data) must be multiples of 512.
        """

        # Check conditions
        if self.masterKey == None:
            raise LuksError("A key has not been unlocked.  Call open_any_key() first.")

        if offset % self.SECTOR_SIZE != 0:
            raise LuksError("offset must be a multiple of %s" % self.SECTOR_SIZE)

        dataLen = len(data)
        if dataLen % self.SECTOR_SIZE != 0:
            raise LuksError("data length must be a multiple of %s" % self.SECTOR_SIZE)

        # Encrypt all the data
        startSector = int(offset / self.SECTOR_SIZE)
        endSector = int((offset + dataLen) / self.SECTOR_SIZE)
        for sector in range(startSector, endSector):
            dslice_sector = sector - startSector
            dslice = data[int(dslice_sector * self.SECTOR_SIZE):int((dslice_sector+1)*self.SECTOR_SIZE)]
            self._encrypt_sector(self.masterKey, self.payloadOffset + sector, sector, dslice)

    def decrypt_data(self, offset, length):
        """Decrypt data from the file.

        Offset is a zero indexed location into the data, and length is
        the ammout of data to return.  offset and length can be any value,
        but multiples of 512 are the most efficient.
        """

        if self.masterKey == None:
            raise LuksError("A key has not been unlocked.  Call open_any_key() first.")

        frontPad = int(offset % self.SECTOR_SIZE)
        startSector = int(math.floor(offset / self.SECTOR_SIZE))
        endSector = int(math.ceil((offset + length) / self.SECTOR_SIZE))
        ret = b""
        for sector in range(startSector, endSector):
            ret += self._decrypt_sector(self.masterKey, self.payloadOffset + sector, sector)

        return ret[frontPad:frontPad+length]

    ##### Private functions

    class _key_block:
        """Internal class, used to store the key information about each key."""

        LUKS_KEY_FORMAT = ">II32sII"

        def load_from_str(self,str):
            """Unpack the key information from a string"""
            self.active, \
            self.passwordIterations, \
            self.passwordSalt, \
            self.keyMaterialOffset, \
            self.stripes =  \
            struct.unpack(self.LUKS_KEY_FORMAT,str)

        def create(self,offset, stripes, disabled):
            """Create a new set of key information.  Called from LuksFile.create()"""
            self.active = disabled
            self.passwordIterations = 0
            self.passwordSalt = b''
            self.keyMaterialOffset = int(offset)
            self.stripes = stripes

        def save(self):
            """Pack the key information into a string"""
            return struct.pack(self.LUKS_KEY_FORMAT, self.active, self.passwordIterations, \
                    self.passwordSalt, self.keyMaterialOffset, self.stripes)

    class _plain_iv_gen:
        """Internal class to represent cbc-plain cipherMode"""

        def set_key(self, key):
            # plain IV generation does not use the key in any way
            pass

        def generate(self, sectorOffset, size):
            istr = struct.pack("<I", sectorOffset)
            return istr + bytes([0] * (size - 4))

    class _essiv_gen:
        """Internal class to represent cbc-essiv:<hash> cipherMode"""

        # essiv mode is defined by
        # SALT=Hash(KEY)
        # IV=E(SALT,sectornumber)
        def __init__(self, str, cipher, luksParent):
            self.hashSpec = str[1:]
            self.cipher = cipher

        def set_key(self, key):
            h = hashlib.new(self.hashSpec, key)
            self.salt = h.digest()
            self.encr = self.cipher.new(self.salt, self.cipher.MODE_ECB)

        def generate(self, sectorOffset, size):
            istr = struct.pack("<I", sectorOffset) + bytes([0] * (size - 4))
            return self.encr.encrypt(istr)

    def _check_cipher(self, cipherName, cipherMode):
        """Internal function to check for a valid cipherName and cipherMode"""
        if cipherName == "aes":
            self.cipher = AES
        elif cipherName == "cast5":
            self.cipher = CAST
        elif cipherName == "blowfish":
            self.cipher = Blowfish
        else:
            raise LuksError("invalid cipher %s" % cipherName)

        # All supported ciphers are block ciphers, so modes are the same (CBC)
        self.mode = self.cipher.MODE_CBC

        if cipherMode == "cbc-plain":
            self.ivGen = self._plain_iv_gen()
        elif cipherMode[:10] == "cbc-essiv:":
            self.ivGen = self._essiv_gen(cipherMode[9:], self.cipher, self)
        else:
            raise LuksError("invalid cipher mode %s" % cipherMode)

        self.cipherName = cipherName
        self.cipherMode = cipherMode

    def _uuidgen(self, rand):
        """Internal function to generate a UUID"""

        # I copied this code (and slightly modified it) from a module written
        # by Denys Duchier http://ofxsuite.berlios.de/uuid.py  (which is under the GPL)

        buf = rand.read(16)
        low,mid,hi_and_version,seq,node = struct.unpack(">IHHH6s",buf)
        seq = (seq & 0x3FFF) | 0x8000
        hi_and_version = (hi_and_version & 0x0FFF) | 0x4000
        uuid = struct.pack(">IHHH6s",low,mid,hi_and_version,seq,node)
        low,mid,hi,seq,b5,b4,b3,b2,b1,b0 = struct.unpack(">IHHHBBBBBB",uuid)
        self.uuid =  b"%08x-%04x-%04x-%02x%02x-%02x%02x%02x%02x%02x%02x" % (low,mid,hi,seq>>8,seq&0xFF,b5,b4,b3,b2,b1,b0)

    def _save_header(self):
        """Internal function to save the header info into the file"""
        str=struct.pack(self.LUKS_FORMAT, self.magic, self.version, self.cipherName.encode(), self.cipherMode.encode(), self.hashSpec.encode(), \
             self.payloadOffset, self.keyBytes, self.mkDigest, self.mkDigestSalt, self.mkDigestIterations, self.uuid)
        self.file.seek(0)
        self.file.write(str)

        for i in range(0, 8):
            self.file.write(self.keys[i].save())

        try:
            self.file.flush()
            os.fsync(self.file.fileno())
        except:
            # We might get an error because self.file.fileno() does not exist on StringIO
            pass

    def _encrypt_sector(self, key, sector, sectorOffset, data):
        """Internal function to encrypt a single sector"""

        if len(data) > self.SECTOR_SIZE:
            raise LuksError("_encrypt_page only accepts data of size <= %i" % self.SECTOR_SIZE)

        # Encrypt the data using the cipher, iv generation, and mode
        IV = self.ivGen.generate(sectorOffset, self.cipher.block_size)
        cipher = self.cipher.new(key, self.mode, IV)
        encrData = cipher.encrypt(data)

        # Write the encrypted data to disk
        self.file.seek(int(sector * self.SECTOR_SIZE))
        self.file.write(encrData)

    def _decrypt_sector(self, key, sector, sectorOffset):
        """Internal function to decrypt a single sector"""

        # Read the ciphertext from disk
        self.file.seek(int(sector * self.SECTOR_SIZE))
        encrData = self.file.read(int(self.SECTOR_SIZE))

        # Decrypt the data using cipher, iv generation, and mode
        IV = self.ivGen.generate(sectorOffset, self.cipher.block_size)
        cipher = self.cipher.new(key, self.mode, IV)
        return cipher.decrypt(encrData)

# The following was copied from the reference implementation of LUKS in cryptsetup-luks-1.0.1 from
# http://luks.endorphin.org/dm-crypt

#define LUKS_CIPHERNAME_L 32
#define LUKS_CIPHERMODE_L 32
#define LUKS_HASHSPEC_L 32
#define LUKS_DIGESTSIZE 20 // since SHA1
#define LUKS_HMACSIZE 32
#define LUKS_SALTSIZE 32
#define LUKS_NUMKEYS 8
#define LUKS_MAGIC_L 6

#/* Actually we need only 37, but we don't want struct autoaligning to kick in */
#define UUID_STRING_L 40

#struct luks_phdr {
#   char        magic[LUKS_MAGIC_L];
#   uint16_t    version;
#   char        cipherName[LUKS_CIPHERNAME_L];
#   char        cipherMode[LUKS_CIPHERMODE_L];
#   char            hashSpec[LUKS_HASHSPEC_L];
#   uint32_t    payloadOffset;
#   uint32_t    keyBytes;
#   char        mkDigest[LUKS_DIGESTSIZE];
#   char        mkDigestSalt[LUKS_SALTSIZE];
#   uint32_t    mkDigestIterations;
#   char            uuid[UUID_STRING_L];
#
#   struct {
#       uint32_t active;
#
#       /* parameters used for password processing */
#       uint32_t passwordIterations;
#       char     passwordSalt[LUKS_SALTSIZE];
#
#       /* parameters used for AF store/load */
#       uint32_t keyMaterialOffset;
#       uint32_t stripes;
#   } keyblock[LUKS_NUMKEYS];
#};

# size is 208 bytes + 48 * LUKS_NUMKEYS  = 592 bytes



# -*- coding: iso-8859-1 -*-
"""
  - Implements the PKCS#5 v2.0: Password-Based Cryptography Standard
    from RSA Laboratories. RFC2898 http://www.rfc-editor.org/rfc/rfc2898.txt

Modifications by John Lenz <lenz@cs.wisc.edu>, April 2006
  + Fix the PBKDFv2 algorithm so it is correct
  + Use Cipher.XOR instead of slow python xor
  + other performance improvements

Original Code written by:

Copyright (C) 2004 - Lars Strand <lars strand at gnist org>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

http://www.gnu.org/copyleft/gpl.html

"""

import struct, string, math, hashlib, hmac # RFC2104

################ PBKDFv2
class PBKDFv2:
    """Implements the PKCS#5 v2.0: Password-Based Cryptography Standard
    from RSA Laboratories. RFC2898

    http://www.rfc-editor.org/rfc/rfc2898.txt
    """

    ################ init
    def __init__(self):

        # length of pseudorandom function: 20 for SHA-1, 16 for MD5
        self.hLen = 20

    ################ makeKey
    def makeKey(self, P, S, c, dkLen, digesttype='sha1'):
        """
           Input:   P         password, an octet string
                    S         salt, an octet string
                    c         iteration count, a positive integer (>1000)
                    dkLen     intended length on octets of the derived key, a positive integer,
                              at most (2^32 - 1) * hLen
                    digesttype Digest used, passed to hmac module

           Output   DK    derived key, a dkLen-octet string
           """

        # do some sanity checks
        try:
            str(P); str(S); int(c); float(dkLen); int(c)
        except:
            print("P = %s, S = %s, c = %s, dkLen = %s:" % (P, S, c, dkLen))
            raise ValueError("ERROR! Input is not correct!")

        # Step 1: if dkLen is larger than maximum possible key - exit
        if dkLen > ((2^32 - 1) * self.hLen):
            maxlength = (2^32 - 1) * self.hLen
            raise ValueError("ERROR! Key is to large! Maxlength is " . str(maxlength))

        # Step 2:
        # Let l be the number of hLen-octet blocks in the derived key, rounding up
        # and let r be the number of octets in the last block
        l = math.ceil(dkLen / float(self.hLen))
        #if (dkLen % float(self.hLen)): l = int(l) + 1 # round up if necessary
        r = dkLen - (l - 1) * self.hLen

        # Step 3:
        # For each block of the derived key, apply the function F to the
        # password P, the salt S, the iteration count c, and the block index
        # to compute the block
        T = b""
        for blockindex in range(int(l)):
            T += self.F(P, S, c, blockindex, digesttype)
        # Step 4 - extract the first dkLen octet to produce a derived key DK
        DK = T[:dkLen]

        # Step 5 - return the derived key DK
        return DK

    ################ F
    def F(self, P, S, c, i, digest):
        """For each block of the derived key, apply this function.

        Notation:
        ||   = concatenation operator
        PRF  = Underlying pseudorandom function

        The function F is defined as the exclusive-or sum of the first c
        iterates if the underlying pseudorandom function PRF applied to
        the password P and the concatenation of the salt S and the block
        index i:

        F(P, S, c, i) = U1 XOR U2 XOR ... XOR Uc

        where

        U1 = PRF(P, S || INT(i)),
        U2 = PRF(P, U1)
        ...
        Uc = PRF(P, Uc-1)
        """

        # The pseudorandom function, PRF, used is HMAC-SHA1 (rfc2104)
        iteration = 1

        # the first iteration; P is the key, and a concatination of
        # S and blocknumber is the message
        istr = struct.pack(">I", i+1)
        PRFMaster = hmac.new(P,digestmod=getattr(hashlib, digest))
        PRF = PRFMaster.copy()
        PRF.update(S)
        PRF.update(istr)
        U = PRF.digest() # the first iteration

        Fbuf = U

        while iteration < c:                  # loop through all iterations
            PRF = PRFMaster.copy()
            PRF.update(U)
            U = PRF.digest()
            Fbuf = self._xor(U, Fbuf)    # XOR this new iteration with the old one
            iteration += 1
        return Fbuf

    ################ xor
    def _xor(self, a, b):
        """Performs XOR on two strings a and b"""

        if len(a) != len(b):
            raise ValueError("ERROR: Strings are of different size! %s %s" % (len(a), len(b)))

        return bytes([x ^ y for (x, y) in zip(a, b)])



#
# Revelation - a password manager for GNOME 2
# http://oss.codepoet.no/revelation/
# $Id$
#
# Module with dialogs
#
#
# Copyright (c) 2003-2006 Erik Grinaker
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from . import config, datahandler, entry, io, ui, util

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import GObject, Gtk, Gio, Gdk
import gettext, urllib.parse

_ = gettext.gettext


EVENT_FILTER        = None
UNIQUE_DIALOGS      = {}


Gtk.rc_parse_string("""
    style "hig" {
        GtkDialog::content-area-border  = 0
        GtkDialog::action-area-border   = 0
    }

    class "GtkDialog" style "hig"
""")


##### EXCEPTIONS #####

class CancelError(Exception):
    "Exception for dialog cancellations"
    pass



##### BASE DIALOGS #####

class Dialog(Gtk.Dialog):
    "Base class for dialogs"

    def __init__(self, parent, title):
        Gtk.Dialog.__init__(
            self, title, parent,
            Gtk.DialogFlags.DESTROY_WITH_PARENT
        )

        self.set_border_width(12)
        self.vbox.set_spacing(12)
        self.set_resizable(False)
        self.set_modal(True)

        self.connect("key-press-event", self.__cb_keypress)


    def __cb_keypress(self, widget, data):
        "Callback for handling key presses"

        # close the dialog on escape
        if data.keyval == Gdk.KEY_Escape:
            self.response(Gtk.ResponseType.CLOSE)
            return True


    def run(self):
        "Runs the dialog"

        self.show_all()

        while True:
            response = Gtk.Dialog.run(self)

            if response == Gtk.ResponseType.NONE:
                continue

            return response



class Popup(Gtk.Window):
    "Base class for popup (frameless) dialogs"

    def __init__(self, widget = None):
        Gtk.Window.__init__(self)
        self.set_decorated(False)

        self.border = Gtk.Frame()
        self.border.set_shadow_type(Gtk.ShadowType.OUT)
        Gtk.Window.add(self, self.border)

        if widget != None:
            self.add(widget)

        self.connect("key-press-event", self.__cb_keypress)


    def __cb_keypress(self, widget, data):
        "Callback for key presses"

        if data.keyval == Gdk.KEY_Escape:
            self.close()


    def add(self, widget):
        "Adds a widget to the window"

        self.border.add(widget)


    def close(self):
        "Closes the dialog"

        self.hide()
        self.emit("closed")
        self.destroy()


    def realize(self):
        "Realizes the popup and displays children"

        Gtk.Window.realize(self)

        for child in self.get_children():
            child.show_all()


    def show(self, x = None, y = None):
        "Show the dialog"

        if x != None and y != None:
            self.move(x, y)

        self.show_all()


GObject.signal_new("closed", Popup, GObject.SignalFlags.ACTION,
                   GObject.TYPE_BOOLEAN, ())



class Utility(Dialog):
    "A utility dialog"

    def __init__(self, parent, title):
        Dialog.__init__(self, parent, title)

        self.set_border_width(12)
        self.vbox.set_spacing(18)

        self.sizegroup = Gtk.SizeGroup(mode=Gtk.SizeGroupMode.HORIZONTAL)


    def add_section(self, title, description = None):
        "Adds an input section to the dialog"

        section = ui.InputSection(title, description, self.sizegroup)
        self.vbox.pack_start(section, True, True, 0)

        return section



class Message(Dialog):
    "A message dialog"

    def __init__(self, parent, title, text, stockimage):
        Dialog.__init__(self, parent, "")

        # hbox with image and contents
        hbox = ui.HBox()
        hbox.set_spacing(12)
        self.vbox.pack_start(hbox, True, True, 0)
        self.vbox.set_spacing(24)

        # set up image
        if stockimage != None:
            image = ui.Image(stockimage, Gtk.IconSize.DIALOG)
            image.set_alignment(0.5, 0)
            hbox.pack_start(image, False, False, 0)

        # set up message
        self.contents = ui.VBox()
        self.contents.set_spacing(10)
        hbox.pack_start(self.contents, True, True, 0)

        label = ui.Label("<span size=\"larger\" weight=\"bold\">%s</span>\n\n%s" % ( util.escape_markup(title), text))
        label.set_alignment(0, 0)
        label.set_selectable(True)
        label.set_max_width_chars(45)
        self.contents.pack_start(label, True, True, 0)


    def run(self):
        "Displays the dialog"

        self.show_all()
        response = Dialog.run(self)
        self.destroy()

        return response



class Error(Message):
    "Displays an error message"

    def __init__(self, parent, title, text):
        Message.__init__(self, parent, title, text, "dialog-error")
        self.add_button(Gtk.STOCK_OK, Gtk.ResponseType.OK)



class Info(Message):
    "Displays an info message"

    def __init__(self, parent, title, text):
        Message.__init__(self, parent, title, text, "dialog-information")
        self.add_button(Gtk.STOCK_OK, Gtk.ResponseType.OK)



class Question(Message):
    "Displays a question"

    def __init__(self, parent, title, text):
        Message.__init__(self, parent, title, text, "dialog-question")
        self.add_button(Gtk.STOCK_OK, Gtk.ResponseType.OK)



class Warning(Message):
    "Displays a warning message"

    def __init__(self, parent, title, text):
        Message.__init__(self, parent, title, text, "dialog-warning")



##### QUESTION DIALOGS #####

class FileChanged(Warning):
    "Notifies about changed file"

    def __init__(self, parent, filename):
        Warning.__init__(
            self, parent, _('File has changed'), _('The current file \'%s\' has changed. Do you want to reload it?') % filename
        )

        self.add_button(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL)
        self.add_button(ui.STOCK_RELOAD, Gtk.ResponseType.OK)
        self.set_default_response(Gtk.ResponseType.OK)


    def run(self):
        "Displays the dialog"

        if Warning.run(self) == Gtk.ResponseType.OK:
            return True

        else:
            raise CancelError


class FileChanges(Warning):
    "Asks to save changes before proceeding"

    def __init__(self, parent, title, text):
        Warning.__init__(self, parent, title, text)

        self.add_button(Gtk.STOCK_DISCARD, Gtk.ResponseType.ACCEPT)
        self.add_button(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL)
        self.add_button(Gtk.STOCK_SAVE, Gtk.ResponseType.OK)
        self.set_default_response(Gtk.ResponseType.OK)


    def run(self):
        "Displays the dialog"

        response = Warning.run(self)

        if response == Gtk.ResponseType.OK:
            return True

        elif response == Gtk.ResponseType.ACCEPT:
            return False

        elif response in ( Gtk.ResponseType.CANCEL, Gtk.ResponseType.CLOSE ):
            raise CancelError



class FileChangesNew(FileChanges):
    "Asks the user to save changes when creating a new file"

    def __init__(self, parent):
        FileChanges.__init__(
            self, parent, _('Save changes to current file?'),
            _('You have made changes which have not been saved. If you create a new file without saving then these changes will be lost.')
        )



class FileChangesOpen(FileChanges):
    "Asks the user to save changes when opening a different file"

    def __init__(self, parent):
        FileChanges.__init__(
            self, parent, _('Save changes before opening?'),
            _('You have made changes which have not been saved. If you open a different file without saving then these changes will be lost.')
        )



class FileChangesQuit(FileChanges):
    "Asks the user to save changes when quitting"

    def __init__(self, parent):
        FileChanges.__init__(
            self, parent, _('Save changes before quitting?'),
            _('You have made changes which have not been saved. If you quit without saving, then these changes will be lost.')
        )

class FileChangesClose(FileChanges):
    "Asks the user to save changes when closing"

    def __init__(self, parent):
        FileChanges.__init__(
            self, parent, _('Save changes before closing?'),
            _('You have made changes which have not been saved. If you close without saving, then these changes will be lost.')
        )


class FileReplace(Warning):
    "Asks for confirmation when replacing a file"

    def __init__(self, parent, file):
        Warning.__init__(
            self, parent, _('Replace existing file?'),
            _('The file \'%s\' already exists - do you wish to replace this file?') % file
        )

        self.add_button(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL)
        self.add_button(ui.STOCK_REPLACE, Gtk.ResponseType.OK)
        self.set_default_response(Gtk.ResponseType.CANCEL)

    def run(self):
        "Displays the dialog"

        if Warning.run(self) == Gtk.ResponseType.OK:
            return True

        else:
            raise CancelError



class FileSaveInsecure(Warning):
    "Asks for confirmation when exporting to insecure file"

    def __init__(self, parent):
        Warning.__init__(
            self, parent, _('Save to insecure file?'),
            _('You have chosen to save your passwords to an insecure (unencrypted) file format - if anyone has access to this file, they will be able to see your passwords.'),
        )

        self.add_button(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL)
        self.add_button(Gtk.STOCK_SAVE, Gtk.ResponseType.OK)
        self.set_default_response(Gtk.ResponseType.CANCEL)


    def run(self):
        "Runs the dialog"

        if Warning.run(self) == Gtk.ResponseType.OK:
            return True

        else:
            raise CancelError



##### FILE SELECTION DIALOGS #####

class FileSelector(Gtk.FileChooserNative):
    "A normal file selector"

    def __init__(self, parent, title=None,
                 action=Gtk.FileChooserAction.OPEN):

        Gtk.FileChooserNative.__init__(self, title=title, action=action)
        self.set_transient_for(parent)
        self.set_local_only(False)
        self.inputsection = None


    def add_widget(self, title, widget):
        "Adds a widget to the file selection"

        if self.inputsection is None:
            self.inputsection = ui.InputSection()
            self.set_extra_widget(self.inputsection)

        self.inputsection.append_widget(title, widget)


    def get_filename(self):
        "Returns the file URI"

        uri = self.get_uri()

        if uri is None:
            return None

        else:
            return io.file_normpath(urllib.parse.unquote(uri))


    def run(self):
        "Displays and runs the file selector, returns the filename"

        response = Gtk.FileChooserNative.run(self)
        filename = self.get_filename()
        self.destroy()

        if response == Gtk.ResponseType.ACCEPT:
            return filename
        else:
            raise CancelError



class ExportFileSelector(FileSelector):
    "A file selector for exporting files"

    def __init__(self, parent):
        FileSelector.__init__(
            self, parent, _('Select File to Export to'),
            Gtk.FileChooserAction.SAVE
        )

        # set up filetype dropdown
        self.dropdown = ui.DropDown()
        self.add_widget(_('Filetype'), self.dropdown)

        for handler in datahandler.get_export_handlers():
            self.dropdown.append_item(handler.name, None, handler)

        for index in range(self.dropdown.get_num_items()):
            if self.dropdown.get_item(index)[2] == datahandler.RevelationXML:
                self.dropdown.set_active(index)


    def run(self):
        "Displays the dialog"

        self.inputsection.show_all()

        if Gtk.FileChooserNative.run(self) == Gtk.ResponseType.ACCEPT:
            filename = self.get_filename()
            handler = self.dropdown.get_active_item()[2]
            self.destroy()

            return filename, handler
        else:
            self.destroy()
            raise CancelError



class ImportFileSelector(FileSelector):
    "A file selector for importing files"

    def __init__(self, parent):
        FileSelector.__init__(
            self, parent, _('Select File to Import'),
            Gtk.FileChooserAction.OPEN
        )

        # set up filetype dropdown
        self.dropdown = ui.DropDown()
        self.add_widget(_('Filetype'), self.dropdown)

        self.dropdown.append_item(_('Automatically detect'))

        for handler in datahandler.get_import_handlers():
            self.dropdown.append_item(handler.name, None, handler)


    def run(self):
        "Displays the dialog"

        self.inputsection.show_all()

        if Gtk.FileChooserNative.run(self) == Gtk.ResponseType.ACCEPT:
            filename = self.get_filename()
            handler = self.dropdown.get_active_item()[2]
            self.destroy()

            return filename, handler
        else:
            self.destroy()
            raise CancelError



class OpenFileSelector(FileSelector):
    "A file selector for opening files"

    def __init__(self, parent):
        FileSelector.__init__(
            self, parent, _('Select File to Open'),
            Gtk.FileChooserAction.OPEN
        )

        filter = Gtk.FileFilter()
        filter.set_name(_('Revelation files'))
        filter.add_mime_type("application/x-revelation")
        self.add_filter(filter)

        filter = Gtk.FileFilter()
        filter.set_name(_('All files'))
        filter.add_pattern("*")
        self.add_filter(filter)



class SaveFileSelector(FileSelector):
    "A file selector for saving files"

    def __init__(self, parent):
        FileSelector.__init__(
            self, parent, _('Select File to Save to'),
            Gtk.FileChooserAction.SAVE
        )

        filter = Gtk.FileFilter()
        filter.set_name(_('Revelation files'))
        filter.add_mime_type("application/x-revelation")
        self.add_filter(filter)

        filter = Gtk.FileFilter()
        filter.set_name(_('All files'))
        filter.add_pattern("*")
        self.add_filter(filter)

        self.set_do_overwrite_confirmation(True)
        self.connect("confirm-overwrite", self.__cb_confirm_overwrite)


    def __cb_confirm_overwrite(self, widget, data = None):
        "Handles confirm-overwrite signals"

        try:
            FileReplace(self, io.file_normpath(self.get_uri())).run()

        except CancelError:
            return Gtk.FileChooserConfirmation.SELECT_AGAIN

        else:
            return Gtk.FileChooserConfirmation.ACCEPT_FILENAME



##### PASSWORD DIALOGS #####

class Password(Message):
    "A base dialog for asking for passwords"

    def __init__(self, parent, title, text, stock = Gtk.STOCK_OK):
        Message.__init__(self, parent, title, text, "dialog-password")

        self.add_button(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL)
        self.add_button(stock, Gtk.ResponseType.OK)
        self.set_default_response(Gtk.ResponseType.OK)

        self.entries = []

        self.sect_passwords = ui.InputSection()
        self.contents.pack_start(self.sect_passwords, True, True, 0)


    def add_entry(self, name, entry = None):
        "Adds a password entry to the dialog"

        if entry is None:
            entry = ui.Entry()
            entry.set_visibility(False)

        self.sect_passwords.append_widget(name, entry)
        self.entries.append(entry)

        return entry


    def run(self):
        "Displays the dialog"

        self.show_all()

        if len(self.entries) > 0:
            self.entries[0].grab_focus()

        return Gtk.Dialog.run(self)



class PasswordChange(Password):
    "A dialog for changing the password"

    def __init__(self, parent, password = None):
        Password.__init__(
            self, parent, _('Enter new password'),
            _('Enter a new password for the current data file. The file must be saved before the new password is applied.'),
            ui.STOCK_PASSWORD_CHANGE
        )

        self.password = password

        if password is not None:
            self.entry_current = self.add_entry(_('Current password'))

        self.entry_new      = self.add_entry(_('New password'), ui.PasswordEntry())
        self.entry_confirm  = self.add_entry(_('Confirm password'))
        self.entry_confirm.autocheck = False


    def run(self):
        "Displays the dialog"

        while True:
            if Password.run(self) != Gtk.ResponseType.OK:
                self.destroy()
                raise CancelError

            elif self.password is not None and self.entry_current.get_text() != self.password:
                Error(self, _('Incorrect password'), _('The password you entered as the current file password is incorrect.')).run()

            elif self.entry_new.get_text() != self.entry_confirm.get_text():
                Error(self, _('Passwords don\'t match'), _('The password and password confirmation you entered does not match.')).run()

            else:
                password = self.entry_new.get_text()

                try:
                    util.check_password(password)

                except ValueError as res:
                    WarnInsecure = Warning(
                        self, _('Use insecure password?'),
                        _('The password you entered is not secure; %s. Are you sure you want to use it?') % str(res).lower()
                    )

                    WarnInsecure.add_button(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL)
                    WarnInsecure.add_button(Gtk.STOCK_OK, Gtk.ResponseType.OK )
                    WarnInsecure.set_default_response(Gtk.ResponseType.CANCEL)

                    response = WarnInsecure.run()
                    if response != Gtk.ResponseType.OK:
                        continue

                self.destroy()
                return password



class PasswordLock(Password):
    "Asks for a password when the file is locked"

    def __init__(self, parent, password):
        Password.__init__(
            self, parent, _('Enter password to unlock file'),
            _('The current file has been locked, please enter the file password to unlock it.'),
            ui.STOCK_UNLOCK
        )

        self.get_widget_for_response(Gtk.ResponseType.CANCEL).set_label(Gtk.STOCK_QUIT)
        self.set_default_response(Gtk.ResponseType.OK)

        self.password = password
        self.entry_password = self.add_entry(_('Password'))


    def run(self):
        "Displays the dialog"

        while True:
            try:
                response = Password.run(self)

                if response == Gtk.ResponseType.CANCEL:
                    raise CancelError

                elif response != Gtk.ResponseType.OK:
                    continue

                elif self.entry_password.get_text() == self.password:
                    break

                else:
                    Error(self, _('Incorrect password'), _('The password you entered was not correct, please try again.')).run()

            except CancelError:
                if self.get_widget_for_response(Gtk.ResponseType.CANCEL).get_property("sensitive") == True:
                    self.destroy()
                    raise

                else:
                    continue

        self.destroy()



class PasswordOpen(Password):
    "Password dialog for opening files"

    def __init__(self, parent, filename):
        Password.__init__(
            self, parent, _('Enter file password'),
            _('The file \'%s\' is encrypted. Please enter the file password to open it.') % filename,
            Gtk.STOCK_OPEN
        )

        self.set_default_response(Gtk.ResponseType.OK)
        self.entry_password = self.add_entry(_('Password'))


    def run(self):
        "Displays the dialog"

        response = Password.run(self)
        password = self.entry_password.get_text()
        self.destroy()

        if response == Gtk.ResponseType.OK:
            return password

        else:
            raise CancelError



class PasswordSave(Password):
    "Password dialog for saving data"

    def __init__(self, parent, filename):
        Password.__init__(
            self, parent, _('Enter password for file'),
            _('Please enter a password for the file \'%s\'. You will need this password to open the file at a later time.') % filename,
            Gtk.STOCK_SAVE
        )

        self.entry_new      = self.add_entry(_('New password'), ui.PasswordEntry())
        self.entry_confirm  = self.add_entry(_('Confirm password'))
        self.entry_confirm.autocheck = False


    def run(self):
        "Displays the dialog"

        while True:
            if Password.run(self) != Gtk.ResponseType.OK:
                self.destroy()
                raise CancelError

            elif self.entry_new.get_text() != self.entry_confirm.get_text():
                Error(self, _('Passwords don\'t match'), _('The passwords you entered does not match.')).run()

            elif len(self.entry_new.get_text()) == 0:
                Error(self, _('No password entered'), _('You must enter a password for the new data file.')).run()

            else:
                password = self.entry_new.get_text()

                try:
                    util.check_password(password)

                except ValueError as res:
                    res = str(res).lower()

                    WarnInsecure = Warning(
                        self, _('Use insecure password?'),
                        _('The password you entered is not secure; %s. Are you sure you want to use it?') % res
                    )

                    WarnInsecure.add_button(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL)
                    WarnInsecure.add_button(Gtk.STOCK_OK, Gtk.ResponseType.OK )
                    WarnInsecure.set_default_response(Gtk.ResponseType.CANCEL)

                    response = WarnInsecure.run()
                    if response != Gtk.ResponseType.OK:
                        continue

                self.destroy()
                return password



##### ENTRY DIALOGS #####

class EntryEdit(Utility):
    "A dialog for editing entries"

    def __init__(self, parent, title, e = None, cfg = None, clipboard = None):
        Utility.__init__(self, parent, title)

        self.add_button(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL)
        if not e:
            self.add_button(ui.STOCK_NEW_ENTRY, Gtk.ResponseType.OK)
        else:
            self.add_button(ui.STOCK_UPDATE, Gtk.ResponseType.OK)
        self.set_default_response(Gtk.ResponseType.OK)

        self.config      = cfg
        self.clipboard   = clipboard
        self.entry_field = {}
        self.fielddata   = {}
        self.widgetdata  = {}

        # set up the ui
        self.sect_meta      = self.add_section(title)
        self.sect_fields    = self.add_section(_('Account Data'))
        self.sect_notes     = self.add_section(_('Notes'))

        self.entry_name = ui.Entry()
        self.entry_name.set_width_chars(50)
        self.entry_name.set_tooltip_text(_('The name of the entry'))
        self.sect_meta.append_widget(_('Name'), self.entry_name)

        self.entry_desc = ui.Entry()
        self.entry_desc.set_tooltip_text(_('A description of the entry'))
        self.sect_meta.append_widget(_('Description'), self.entry_desc)

        self.dropdown = ui.EntryDropDown()
        self.dropdown.connect("changed", lambda w: self.__setup_fieldsect(self.dropdown.get_active_type()().fields))
        eventbox = ui.EventBox(self.dropdown)
        eventbox.set_tooltip_text(_('The type of entry'))
        self.sect_meta.append_widget(_('Type'), eventbox)

        self.entry_notes = ui.EditableTextView()
        self.entry_notes.set_tooltip_text(_('Notes for the entry'))
        self.sect_notes.append_widget(None, self.entry_notes)

        # populate the dialog with data
        self.set_entry(e)


    def __setup_fieldsect(self, fields):
        "Generates a field section based on a field list"

        # store current field values
        for fieldtype, fieldentry in self.entry_field.items():
            self.fielddata[fieldtype] = fieldentry.get_text()

        self.entry_field = {}
        self.sect_fields.clear()

        # generate field entries
        for field in fields:

            if type(field) in self.widgetdata:
                userdata = self.widgetdata[type(field)]

            elif field.datatype == entry.DATATYPE_PASSWORD:
                userdata = self.clipboard

            else:
                userdata = None

            fieldentry = ui.generate_field_edit_widget(field, self.config, userdata)
            self.entry_field[type(field)] = fieldentry

            if type(field) in self.fielddata:
                fieldentry.set_text(self.fielddata[type(field)])

            if hasattr(fieldentry, "set_tooltip_text"):
                fieldentry.set_tooltip_text(field.description)

            elif hasattr(fieldentry, "entry") == True:
                fieldentry.entry.set_tooltip_text(field.description)

            self.sect_fields.append_widget(field.name, fieldentry)


        # show widgets
        self.sect_fields.show_all()


    def get_entry(self):
        "Generates an entry from the dialog contents"

        e = self.dropdown.get_active_type()()

        e.name = self.entry_name.get_text()
        e.description = self.entry_desc.get_text()
        e.notes = self.entry_notes.get_text()

        for field in e.fields:
            field.value = self.entry_field[type(field)].get_text()

        return e


    def run(self):
        "Displays the dialog"

        while True:
            self.show_all()

            if Utility.run(self) == Gtk.ResponseType.OK:
                e = self.get_entry()

                if e.name == "":
                    Error(self, _('Name not entered'), _('You must enter a name for the account')).run()
                    continue

                self.destroy()
                return e

            else:
                self.destroy()
                raise CancelError


    def set_entry(self, e):
        "Sets an entry for the dialog"

        e = e is not None and e.copy() or entry.GenericEntry()

        self.entry_name.set_text(e.name)
        self.entry_desc.set_text(e.description)
        self.entry_notes.set_text(e.notes)
        self.dropdown.set_active_type(type(e))

        for field in e.fields:
            self.entry_field[type(field)].set_text(field.value or "")


    def set_fieldwidget_data(self, fieldtype, userdata):
        "Sets user data for fieldwidget"

        self.widgetdata[fieldtype] = userdata

        if fieldtype == entry.UsernameField and entry.UsernameField in self.entry_field:
            self.entry_field[entry.UsernameField].set_values(userdata)



class EntryRemove(Warning):
    "Asks for confirmation when removing entries"

    def __init__(self, parent, entries):

        if len(entries) > 1:
            title   = _('Really remove the %i selected entries?') % len(entries)
            text    = _('By removing these entries you will also remove any entries they may contain.')

        elif type(entries[0]) == entry.FolderEntry:
            title   = _('Really remove folder \'%s\'?') % entries[0].name
            text    = _('By removing this folder you will also remove all accounts and folders it contains.')

        else:
            title   = _('Really remove account \'%s\'?') % entries[0].name
            text    = _('Please confirm that you wish to remove this account.')


        Warning.__init__(self, parent, title, text)

        self.add_button(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL)
        self.add_button(Gtk.STOCK_REMOVE, Gtk.ResponseType.OK)
        self.set_default_response(Gtk.ResponseType.CANCEL)


    def run(self):
        "Displays the dialog"

        if Warning.run(self) == Gtk.ResponseType.OK:
            return True

        else:
            raise CancelError



class FolderEdit(Utility):
    "Dialog for editing a folder"

    def __init__(self, parent, title, e = None):
        Utility.__init__(self, parent, title)

        self.add_button(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL)
        if not e:
            self.add_button(ui.STOCK_NEW_FOLDER, Gtk.ResponseType.OK)
        else:
            self.add_button(ui.STOCK_UPDATE, Gtk.ResponseType.OK)
        self.set_default_response(Gtk.ResponseType.OK)

        # set up the ui
        self.sect_folder    = self.add_section(title)

        self.entry_name = ui.Entry()
        self.entry_name.set_width_chars(25)
        self.entry_name.set_tooltip_text(_('The name of the folder'))
        self.sect_folder.append_widget(_('Name'), self.entry_name)

        self.entry_desc = ui.Entry()
        self.entry_desc.set_tooltip_text(_('A description of the folder'))
        self.sect_folder.append_widget(_('Description'), self.entry_desc)

        # populate the dialog with data
        self.set_entry(e)


    def get_entry(self):
        "Generates an entry from the dialog contents"

        e = entry.FolderEntry()
        e.name = self.entry_name.get_text()
        e.description = self.entry_desc.get_text()

        return e


    def run(self):
        "Displays the dialog"

        while True:
            self.show_all()

            if Utility.run(self) == Gtk.ResponseType.OK:
                e = self.get_entry()

                if e.name == "":
                    Error(self, _('Name not entered'), _('You must enter a name for the folder')).run()
                    continue

                self.destroy()
                return e

            else:
                self.destroy()
                raise CancelError


    def set_entry(self, e):
        "Sets an entry for the dialog"

        if e is None:
            return

        self.entry_name.set_text(e.name)
        self.entry_desc.set_text(e.description)



##### MISCELLANEOUS DIALOGS #####

class About(Gtk.AboutDialog):
    "About dialog"

    def __init__(self, parent):
        Gtk.AboutDialog.__init__(self)

        if isinstance(parent, Gtk.Window):
            self.set_transient_for(parent)

        self.set_name(config.APPNAME)
        self.set_version(config.VERSION)
        self.set_copyright(config.COPYRIGHT)
        self.set_comments(('"%s"\n\n' + _('A password manager for the GNOME desktop')) % config.RELNAME)
        self.set_license(config.LICENSE)
        self.set_website(config.URL)
        self.set_authors(config.AUTHORS)
        self.set_artists(config.ARTISTS)


    def run(self):
        "Displays the dialog"

        self.show_all()
        Gtk.AboutDialog.run(self)

        self.destroy()



class Exception(Error):
    "Displays a traceback for an unhandled exception"

    def __init__(self, parent, traceback):
        Error.__init__(
            self, parent, _('Unknown error'),
            _('An unknown error occured. Please report the text below to the Revelation developers, along with what you were doing that may have caused the error. You may attempt to continue running Revelation, but it may behave unexpectedly.')
        )

        self.add_button(Gtk.STOCK_QUIT, Gtk.ResponseType.CANCEL )
        self.add_button(ui.STOCK_CONTINUE, Gtk.ResponseType.OK )

        textview = ui.TextView(None, traceback)
        scrolledwindow = ui.ScrolledWindow(textview)
        scrolledwindow.set_size_request(-1, 120)

        self.contents.pack_start(scrolledwindow, True, True, 0)


    def run(self):
        "Runs the dialog"

        return Error.run(self) == Gtk.ResponseType.OK



class PasswordChecker(Utility):
    "A password checker dialog"

    def __init__(self, parent, cfg = None, clipboard = None):
        Utility.__init__(self, parent, _('Password Checker'))

        self.add_button(Gtk.STOCK_CLOSE, Gtk.ResponseType.CLOSE)
        self.set_default_response(Gtk.ResponseType.CLOSE)

        self.cfg = cfg
        self.set_modal(False)
        self.set_size_request(300, -1)

        self.section = self.add_section(_('Password Checker'))

        self.entry = ui.PasswordEntry(None, cfg, clipboard)
        self.entry.autocheck = False
        self.entry.set_width_chars(40)
        self.entry.connect("changed", self.__cb_changed)
        self.entry.set_tooltip_text(_('Enter a password to check'))
        self.section.append_widget(_('Password'), self.entry)

        self.result = ui.ImageLabel(_('Enter a password to check'), ui.STOCK_UNKNOWN, ui.ICON_SIZE_HEADLINE)
        self.result.set_tooltip_text(_('The result of the check'))
        self.section.append_widget(None, self.result)

        self.connect("response", self.__cb_response)


    def __cb_changed(self, widget, data = None):
        "Callback for entry changes"

        password = self.entry.get_text()

        try:
            if len(password) == 0:
                icon    = ui.STOCK_UNKNOWN
                result  = _('Enter a password to check')

            else:
                util.check_password(password)
                icon    = ui.STOCK_PASSWORD_STRONG
                result  = _('The password seems good')

        except ValueError as reason:
            icon    = ui.STOCK_PASSWORD_WEAK
            result = _('The password %s') % str(reason)

        self.result.set_text(result)
        self.result.set_stock(icon, ui.ICON_SIZE_HEADLINE)


    def __cb_response(self, widget, response):
        "Callback for response"

        self.destroy()


    def run(self):
        "Displays the dialog"

        self.show_all()

        # for some reason, Gtk crashes on close-by-escape
        # if we don't do this
        self.get_widget_for_response(Gtk.ResponseType.CLOSE).grab_focus()
        self.entry.grab_focus()



class PasswordGenerator(Utility):
    "A password generator dialog"

    def __init__(self, parent, cfg, clipboard = None):
        Utility.__init__(self, parent, _('Password Generator'))

        self.add_button(Gtk.STOCK_CLOSE, Gtk.ResponseType.CLOSE)
        self.add_button(ui.STOCK_GENERATE, Gtk.ResponseType.OK)

        self.config = cfg
        self.set_modal(False)

        self.section = self.add_section(_('Password Generator'))

        self.entry = ui.PasswordEntry(None, cfg, clipboard)
        self.entry.autocheck = False
        self.entry.set_editable(False)
        self.entry.set_tooltip_text(_('The generated password'))
        self.section.append_widget(_('Password'), self.entry)

        self.spin_pwlen = ui.SpinEntry()
        self.spin_pwlen.set_range(4, 256)
        self.config.bind("passwordgen-length", self.spin_pwlen, "value", Gio.SettingsBindFlags.DEFAULT)
        self.spin_pwlen.set_tooltip_text(_('The number of characters in generated passwords - 8 or more are recommended'))
        self.section.append_widget(_('Length'), self.spin_pwlen)

        self.check_punctuation_chars = ui.CheckButton(_('Use punctuation characters for passwords'))
        if self.config.get_int("passwordgen-length"):
            self.check_punctuation_chars.set_sensitive(True)
        self.config.bind("passwordgen-punctuation", self.check_punctuation_chars, "active", Gio.SettingsBindFlags.DEFAULT)
        self.check_punctuation_chars.set_tooltip_text(_('When passwords are generated, use punctuation characters like %, =, { or .'))
        self.section.append_widget(None, self.check_punctuation_chars)

        self.connect("response", self.__cb_response)


    def __cb_response(self, widget, response):
        "Callback for dialog responses"

        if response == Gtk.ResponseType.OK:
            self.entry.set_text(util.generate_password(self.spin_pwlen.get_value(), self.check_punctuation_chars.get_active()))

        else:
            self.destroy()


    def run(self):
        "Displays the dialog"

        self.show_all()
        self.get_widget_for_response(Gtk.ResponseType.OK).grab_focus()



##### FUNCTIONS #####

def create_unique(dialog, *args):
    "Creates a unique dialog"

    if present_unique(dialog) == True:
        return get_unique(dialog)

    else:
        UNIQUE_DIALOGS[dialog] = dialog(*args)
        UNIQUE_DIALOGS[dialog].connect("destroy", lambda w: remove_unique(dialog))

        return UNIQUE_DIALOGS[dialog]


def get_unique(dialog):
    "Returns a unique dialog"

    if unique_exists(dialog) == True:
        return UNIQUE_DIALOGS[dialog]

    else:
        return None


def present_unique(dialog):
    "Presents a unique dialog, if it exists"

    if unique_exists(dialog) == True:
        get_unique(dialog).present()

        return True

    else:
        return False


def remove_unique(dialog):
    "Removes a unique dialog"

    if unique_exists(dialog):
        UNIQUE_DIALOGS[dialog] = None


def run_unique(dialog, *args):
    "Runs a unique dialog"

    if present_unique(dialog) == True:
        return None

    else:
        d = create_unique(dialog, *args)

        return d.run()


def unique_exists(dialog):
    "Checks if a unique dialog exists"

    return dialog in UNIQUE_DIALOGS and UNIQUE_DIALOGS[dialog] != None




#
# Revelation - a password manager for GNOME 2
# http://oss.codepoet.no/revelation/
# $Id$
#
# Module containing entry information
#
#
# Copyright (c) 2003-2006 Erik Grinaker
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from revelation import ui

import copy, gettext, time

_ = gettext.gettext


DATATYPE_FILE       = "file"
DATATYPE_EMAIL      = "email"
DATATYPE_PASSWORD   = "password"
DATATYPE_STRING     = "string"
DATATYPE_URL        = "url"



class EntryFieldError(Exception):
    "Exception for invalid entry fields"
    pass


class EntryTypeError(Exception):
    "Exception for invalid entry types"
    pass



class Entry(object):
    "An entry object"

    id      = None
    typename    = ""
    icon        = None

    def __init__(self):
        self.name       = ""
        self.description    = ""
        self.notes      = ""
        self.updated        = int(time.time())
        self.fields     = []


    def __getitem__(self, key):
        return self.get_field(key).value


    def __setitem__(self, key, value):
        self.get_field(key).value = value


    def convert_generic(self):
        "Creates a GenericEntry with the data from this entry"

        generic = GenericEntry()
        generic.name = self.name
        generic.description = self.description
        generic.notes = self.notes
        generic.updated = self.updated

        # do direct field copies
        for field in generic.fields:
            if self.has_field(type(field)):
                field.value = self[type(field)]

        # handle special conversions
        if type(self) == CreditcardEntry:
            generic[UsernameField] = self[CardnumberField]
            generic[PasswordField] = self[PINField]

        elif type(self) == CryptoKeyEntry:
            generic[UsernameField] = self[KeyfileField]

        elif type(self) == DatabaseEntry:
            if self[DatabaseField] != "":
                generic[HostnameField] = self[DatabaseField] + "@" + self[HostnameField]

        elif type(self) == DoorEntry:
            generic[PasswordField] = self[CodeField]
            generic[HostnameField] = self[LocationField]

        elif type(self) == FTPEntry:
            if self[PortField] not in ( "", "21" ):
                generic[HostnameField] += ":" + self[PortField]

        elif type(self) == PhoneEntry:
            generic[UsernameField] = self[PhonenumberField]
            generic[PasswordField] = self[PINField]

        elif type(self) == WebEntry:
            generic[HostnameField] = self[URLField]


        return generic


    def copy(self):
        "Create a copy of the entry"

        return copy.deepcopy(self)


    def get_field(self, fieldtype):
        "Get one of the entrys fields"

        for field in self.fields:
            if type(field) == fieldtype:
                return field

        else:
            raise EntryFieldError


    def has_field(self, fieldtype):
        "Check if the entry has a field"

        try:
            self.get_field(fieldtype)
            return True

        except EntryFieldError:
            return False


    def mirror(self, entry):
        "Makes this entry mirror a different entry (same data)"

        if type(self) != type(entry):
            raise EntryTypeError

        self.name       = entry.name
        self.description    = entry.description
        self.notes      = entry.notes
        self.updated        = entry.updated

        for field in entry.fields:
            self[type(field)] = field.value



class FolderEntry(Entry):

    def __init__(self):
        Entry.__init__(self)

        self.id     = "folder"
        self.typename   = _('Folder')
        self.icon   = ui.STOCK_ENTRY_FOLDER
        self.openicon   = ui.STOCK_ENTRY_FOLDER_OPEN



class CreditcardEntry(Entry):

    def __init__(self):
        Entry.__init__(self)

        self.id     = "creditcard"
        self.typename   = _('Creditcard')
        self.icon   = ui.STOCK_ENTRY_CREDITCARD

        self.fields = [
            CardtypeField(),
            CardnumberField(),
            ExpirydateField(),
            CCVField(),
            PINField()
        ]



class CryptoKeyEntry(Entry):

    def __init__(self):
        Entry.__init__(self)

        self.id     = "cryptokey"
        self.typename   = _('Crypto Key')
        self.icon   = ui.STOCK_ENTRY_CRYPTOKEY

        self.fields = [
            HostnameField(),
            CertificateField(),
            KeyfileField(),
            PasswordField()
        ]



class DatabaseEntry(Entry):

    def __init__(self):
        Entry.__init__(self)

        self.id     = "database"
        self.typename   = _('Database')
        self.icon   = ui.STOCK_ENTRY_DATABASE

        self.fields = [
            HostnameField(),
            UsernameField(),
            PasswordField(),
            DatabaseField()
        ]



class DoorEntry(Entry):

    def __init__(self):
        Entry.__init__(self)

        self.id     = "door"
        self.typename   = _('Door lock')
        self.icon   = ui.STOCK_ENTRY_DOOR

        self.fields = [
            LocationField(),
            CodeField()
        ]



class EmailEntry(Entry):

    def __init__(self):
        Entry.__init__(self)

        self.id     = "email"
        self.typename   = _('Email')
        self.icon   = ui.STOCK_ENTRY_EMAIL

        self.fields = [
            EmailField(),
            HostnameField(),
            UsernameField(),
            PasswordField()
        ]



class FTPEntry(Entry):

    def __init__(self):
        Entry.__init__(self)

        self.id     = "ftp"
        self.typename   = _('FTP')
        self.icon   = ui.STOCK_ENTRY_FTP

        self.fields = [
            HostnameField(),
            PortField(),
            UsernameField(),
            PasswordField()
        ]



class GenericEntry(Entry):

    def __init__(self):
        Entry.__init__(self)

        self.id     = "generic"
        self.typename   = _('Generic')
        self.icon   = ui.STOCK_ENTRY_GENERIC

        self.fields = [
            HostnameField(),
            UsernameField(),
            PasswordField()
        ]



class PhoneEntry(Entry):

    def __init__(self):
        Entry.__init__(self)

        self.id     = "phone"
        self.typename   = _('Phone')
        self.icon   = ui.STOCK_ENTRY_PHONE

        self.fields = [
            PhonenumberField(),
            PINField()
        ]



class ShellEntry(Entry):

    def __init__(self):
        Entry.__init__(self)

        self.id     = "shell"
        self.typename   = _('Shell')
        self.icon   = ui.STOCK_ENTRY_SHELL

        self.fields = [
            HostnameField(),
            DomainField(),
            UsernameField(),
            PasswordField()
        ]

class RemoteDesktopEntry(Entry):

    def __init__(self):
        Entry.__init__(self)

        self.id     = "remotedesktop"
        self.typename   = _('Remote Desktop')
        self.icon   = ui.STOCK_ENTRY_REMOTEDESKTOP

        self.fields = [
            HostnameField(),
            PortField(),
            UsernameField(),
            PasswordField()
        ]


class WebEntry(Entry):

    def __init__(self):
        Entry.__init__(self)

        self.id     = "website"
        self.typename   = _('Website')
        self.icon   = ui.STOCK_ENTRY_WEBSITE

        self.fields = [
            URLField(),
            UsernameField(),
            EmailField(),
            PasswordField()
        ]

class VNCEntry(Entry):

    def __init__(self):
        Entry.__init__(self)

        self.id     = "vnc"
        self.typename   = _('VNC')
        self.icon   = ui.STOCK_ENTRY_REMOTEDESKTOP

        self.fields = [
            HostnameField(),
            PortField(),
            UsernameField(),
            PasswordField()
        ]

ENTRYLIST = [
    FolderEntry,
    CreditcardEntry,
    CryptoKeyEntry,
    DatabaseEntry,
    DoorEntry,
    EmailEntry,
    FTPEntry,
    GenericEntry,
    PhoneEntry,
    ShellEntry,
    RemoteDesktopEntry,
    VNCEntry,
    WebEntry
]



class Field(object):
    "An entry field object"

    id      = None
    name        = ""
    description = ""
    symbol      = None

    datatype    = None
    value       = None

    def __init__(self, value = ""):
        self.value = value


    def __str__(self):
        return self.value is not None and self.value or ""



class CardnumberField(Field):

    id      = "creditcard-cardnumber"
    symbol      = "N"
    datatype    = DATATYPE_STRING

    def __init__(self, value = ""):
        self.name       = _('Card number')
        self.description    = _('The number of a creditcard, usually a 16-digit number')



class CardtypeField(Field):

    id      = "creditcard-cardtype"
    symbol      = "C"
    datatype    = DATATYPE_STRING

    def __init__(self, value = ""):
        self.name       = _('Card type')
        self.description    = _('The type of creditcard, like MasterCard or VISA')



class CCVField(Field):

    id      = "creditcard-ccv"
    symbol      = "V"
    datatype    = DATATYPE_STRING

    def __init__(self, value = ""):
        self.name       = _('CCV number')
        self.description    = _('A Credit Card Verification number, normally a 3-digit code found on the back of a card')



class CertificateField(Field):

    id      = "generic-certificate"
    symbol      = "x"
    datatype    = DATATYPE_FILE

    def __init__(self, value = ""):
        self.name       = _('Certificate')
        self.description    = _('A certificate, such as an X.509 SSL Certificate')



class CodeField(Field):

    id      = "generic-code"
    symbol      = "c"
    datatype    = DATATYPE_PASSWORD

    def __init__(self, value = ""):
        self.name       = _('Code')
        self.description    = _('A code used to provide access to something')



class DatabaseField(Field):

    id      = "generic-database"
    symbol      = "D"
    datatype    = DATATYPE_STRING

    def __init__(self, value = ""):
        self.name       = _('Database')
        self.description    = _('A database name')



class DomainField(Field):

    id      = "generic-domain"
    symbol      = "d"
    datatype    = DATATYPE_STRING

    def __init__(self, value = ""):
        self.name       = _('Domain')
        self.description    = _('An Internet or logon domain, like organization.org or a Windows logon domain')



class EmailField(Field):

    id      = "generic-email"
    symbol      = "e"
    datatype    = DATATYPE_EMAIL

    def __init__(self, value = ""):
        self.name       = _('Email')
        self.description    = _('An email address')



class ExpirydateField(Field):

    id      = "creditcard-expirydate"
    symbol      = "E"
    datatype    = DATATYPE_STRING

    def __init__(self, value = ""):
        self.name       = _('Expiry date')
        self.description    = _('The month that the credit card validity expires')



class HostnameField(Field):

    id      = "generic-hostname"
    symbol      = "h"
    datatype    = DATATYPE_STRING

    def __init__(self, value = ""):
        self.name       = _('Hostname')
        self.description    = _('The name of a computer, like computer.domain.com or MYCOMPUTER')



class KeyfileField(Field):

    id      = "generic-keyfile"
    symbol      = "f"
    datatype    = DATATYPE_FILE

    def __init__(self, value = ""):
        self.name       = _('Key File')
        self.description    = _('A key file, used for authentication for example via ssh or to encrypt X.509 certificates')



class LocationField(Field):

    id      = "generic-location"
    symbol      = "L"
    datatype    = DATATYPE_STRING

    def __init__(self, value = ""):
        self.name       = _('Location')
        self.description    = _('A physical location, like office entrance')



class PasswordField(Field):

    id      = "generic-password"
    symbol      = "p"
    datatype    = DATATYPE_PASSWORD

    def __init__(self, value = ""):
        self.name       = _('Password')
        self.description    = _('A secret word or character combination used for proving you have access')



class PhonenumberField(Field):

    id      = "phone-phonenumber"
    symbol      = "n"
    datatype    = DATATYPE_STRING

    def __init__(self, value = ""):
        self.name       = _('Phone number')
        self.description    = _('A telephone number')



class PINField(Field):

    id      = "generic-pin"
    symbol      = "P"
    datatype    = DATATYPE_PASSWORD

    def __init__(self, value = ""):
        self.name       = _('PIN')
        self.description    = _('A Personal Identification Number, a numeric code used for credit cards, phones etc')



class PortField(Field):

    id      = "generic-port"
    symbol      = "o"
    datatype    = DATATYPE_STRING

    def __init__(self, value = ""):
        self.name       = _('Port number')
        self.description    = _('A network port number, used to access network services directly')



class URLField(Field):

    id      = "generic-url"
    symbol      = "U"
    datatype    = DATATYPE_URL

    def __init__(self, value = ""):
        self.name       = _('URL')
        self.description    = _('A Uniform Resource Locator, such as a web-site address')



class UsernameField(Field):

    id      = "generic-username"
    symbol      = "u"
    datatype    = DATATYPE_STRING

    def __init__(self, value = ""):
        self.name       = _('Username')
        self.description    = _('A name or other identification used to identify yourself')



FIELDLIST = [
    CardnumberField,
    CardtypeField,
    CCVField,
    CertificateField,
    CodeField,
    DatabaseField,
    DomainField,
    EmailField,
    ExpirydateField,
    HostnameField,
    KeyfileField,
    LocationField,
    PasswordField,
    PhonenumberField,
    PINField,
    PortField,
    URLField,
    UsernameField
]




#
# Revelation - a password manager for GNOME 2
# http://oss.codepoet.no/revelation/
# $Id$
#
# Module for exporting to XHTML files
#
#
# Copyright (c) 2003-2006 Erik Grinaker
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from . import base
from revelation import config, entry

import gettext, time

_ = gettext.gettext


IMAGEPATH   = "https://revelation.olasagasti.info/static"


class XHTML(base.DataHandler):
    "Data handler for XHTML export"

    name        = "XHTML / CSS"
    importer    = False
    exporter    = True
    encryption  = False


    def __init__(self):
        base.DataHandler.__init__(self)


    def __generate_css(self):
        "Generates a CSS for the XHTML document"

        css = """
        /* containers */
        body {
            margin: 20px 10px;
            padding: 0px;

            font-family: sans-serif;
            font-size: 10pt;
            line-height: 1.3;

            color: #333333;
            background-color: #ffffff;
        }

        div {
            margin: 0px;
            padding: 0px;
        }

        p {
            margin: 0px 0px 5px 0px;
            padding: 0px;
        }

        table {
            font-family: sans-serif;
            font-size: 10pt;
            line-height: 1.3;

            margin: 0px;
            border-collapse: collapse;

            color: #333333;
            background-color: transparent;
        }

        td {
            text-align: left;
            vertical-align: top;
        }

        th {
            text-align: left;
            vertical-align: top;

            font-weight: bold;
            color: #000000;
        }



        /* headings */
        h1, h2, h3 {
            margin: 0px;
            padding: 0px;
            line-height: 1;
            font-weight: bold;
            color: #000000;
        }

        h1 {
            margin-bottom: 20px;
            font-size: 16pt;
        }

        h2 {
            margin-bottom: 5px;
            font-size: 14pt;
        }

        h3 {
            font-size: 10pt;
        }



        /* content */
        a {
            color: #3333cc;
            text-decoration: none;
        }

        a:hover {
            text-decoration: underline;
        }

        img {
            margin: 0px;
            padding: 0px;
            border: none;
        }

        strong {
            color: #000000;
        }



        /* the content area */
        #content {
            margin-right: 220px;
        }



        /* the sidebar */
        #sidebar {
            width: 200px;
            float: right;
        }

        #sidebar h2 {
            margin: 20px 0px 10px 0px;
            padding: 0px 0px 5px 0px;
            border-bottom: 1px dashed #3366cc;
            text-transform: lowercase;
        }

        #sidebar h2 img.icon {
            width: 24px;
            height: 24px;
            margin-right: 5px;
            vertical-align: middle;
        }

        #sidebar h3 {
            margin-bottom: 2px;
            text-transform: lowercase;
        }

        #sidebar p {
            margin: 0px 5px 10px 15px;
        }

        #sidebar ul {
            list-style: none;
            padding: 0px;
            margin: 0px 0px 10px 0px;
        }

        #sidebar ul ul {
            margin: 0px 0px 0px 15px;
        }

        #sidebar ul.accountlist {
            margin-left: 15px;
        }



        /* the entrylist */
        #entrylist {
            width: 99%;
            padding: 0px;
            margin: 0px;
        }



        /* the footer */
        #footer {
            margin: 50px 220px 10px 0px;
            font-size: 8pt;
            text-align: center;
            color: #aaaaaa;
        }

        #footer a {
            color: #8888ee;
        }



        /* folders */
        li.folder {
            width: 100%;
            list-style-type: none;
        }

        li.folder .folder-data {
            margin: 0px 0px 10px 0px;
            padding: 2px 5px 3px 5px;

            border: 1px solid #3366cc;
            background-color: #e5ecf9;
        }

        li.folder .folder-data h2 {
            margin: 0px;
        }

        li.folder .folder-data p {
            margin: 0px;
        }



        /* accounts  */
        li.account {
            list-style-type: none;
            width: 350px;
            margin: 0px 0px 10px 0px;
            border: 1px solid #3366cc;
        }

        li.account .heading {
            padding: 2px 5px;
            background-color: #e5ecf9;
            border-bottom: 1px solid #3366cc;
        }

        li.account .heading h2 {
            margin: 0px 0px 1px 0px;
        }

        li.account .heading img {
            float: right;
            width: 24px;
            height: 24px;
        }

        li.account .heading .description {
            margin: 0px;
        }

        li.account .heading .type {
            color: #000000;
            font-weight: bold;
        }

        li.account .data {
            padding: 5px;
        }

        li.account .data .fields {
            margin: 0px 0px 3px 0px;
        }

        li.account .data .fields td {
            padding: 1px 0px;
        }

        li.account .data .fields th {
            padding: 1px 5px 1px 0px;
        }

        li.account .updated {
            font-size: 8pt;
            text-align: right;
            margin: 0px;
        }
"""

        return css


    def __generate_entry(self, entrystore, iter, depth = 0):
        "Generates xhtml for an entry"

        tabs = "\t" * (depth + 2)
        e = entrystore.get_entry(iter)

        if e is not None:
            e.path = self.__get_entryid(entrystore, iter)


        xhtml = ""

        if e is None:

            xhtml += "<div id=\"content\">\n"
            xhtml += "  <ul id=\"entrylist\">\n"

            for i in range(entrystore.iter_n_children(iter)):
                child = entrystore.iter_nth_child(iter, i)
                xhtml += self.__generate_entry(entrystore, child, depth)

            xhtml += "  </ul>\n"
            xhtml += "</div>\n"


        elif type(e) == entry.FolderEntry:
            xhtml += tabs + "<li class=\"folder\" id=\"%s\">\n" % e.path
            xhtml += tabs + "   <div class=\"folder-data\">\n"
            xhtml += tabs + "       <h2>%s</h2>\n" % e.name

            if e.description != "":
                xhtml += tabs + "       <p class=\"description\">%s</p>\n" % e.description

            xhtml += tabs + "   </div>\n"
            xhtml += tabs + "\n"
            xhtml += tabs + "   <ul>\n"

            for i in range(entrystore.iter_n_children(iter)):
                child = entrystore.iter_nth_child(iter, i)
                xhtml += self.__generate_entry(entrystore, child, depth + 2)

            xhtml += tabs + "   </ul>\n"
            xhtml += tabs + "</li>\n"
            xhtml += tabs + "\n"


        else:

            xhtml += tabs + "<li class=\"account\" id=\"%s\">\n" % e.path
            xhtml += tabs + "   <div class=\"heading\">\n"
            xhtml += tabs + "       <img src=\"%s/entry/%s.png\" alt=\"%s\" />\n" % ( IMAGEPATH, e.id, e.typename )
            xhtml += tabs + "       <h2>%s</h2>\n" % e.name
            xhtml += tabs + "       <p class=\"description\"><span class=\"type\">%s</span>%s</p>\n" % ( e.typename + (e.description != "" and "; " or ""), e.description )
            xhtml += tabs + "   </div>\n"
            xhtml += tabs + "\n"

            xhtml += tabs + "   <div class=\"data\">\n"

            fields = []
            for field in e.fields:
                if field.value != "":
                    fields.append(field)

            if len(fields) > 0:
                xhtml += tabs + "       <table class=\"fields\">\n"

                for field in fields:
                    xhtml += tabs + "           <tr>\n"
                    xhtml += tabs + "               <th>" + field.name + ":</th>\n"
                    xhtml += tabs + "               <td>" + field.value + "</td>\n"
                    xhtml += tabs + "           </tr>\n"

                xhtml += tabs + "       </table>\n"

            xhtml += tabs + "       <p class=\"updated\">Updated " + time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(e.updated)) + "</p>\n"
            xhtml += tabs + "   </div>\n"
            xhtml += tabs + "</li>\n"
            xhtml += tabs + "\n"

        return xhtml


    def __generate_footer(self):
        "Generates an xhtml footer"

        xhtml = ""
        xhtml += "<div id=\"footer\">\n"
        xhtml += "  Generated by <a href=\"%s\">%s %s</a>" % ( config.URL, config.APPNAME, config.VERSION )
        xhtml += "</div>\n"
        xhtml += "</body>\n"
        xhtml += "</html>\n"

        return xhtml


    def __generate_header(self):
        "Generates an xhtml header"

        xhtml = ""
        xhtml += "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Strict//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd\">\n"
        xhtml += "<html xmlns=\"http://www.w3.org/1999/xhtml\">\n"
        xhtml += "<head>\n"
        xhtml += "  <title>Revelation account list</title>\n"
        xhtml += "  <meta http-equiv=\"Content-Type\" content=\"text/html; charset=ISO-8859-1\" />\n"
        xhtml += "\n"
        xhtml += "  <style type=\"text/css\">\n"
        xhtml += self.__generate_css()
        xhtml += "  </style>\n"
        xhtml += "</head>\n"
        xhtml += "\n"
        xhtml += "<body>\n"
        xhtml += "<h1>" + _('Revelation account list') + "</h1>\n"
        xhtml += "\n"

        return xhtml


    def __generate_sidebar(self, entrystore):
        "Generates a sidebar"

        xhtml = ""
        xhtml += "<div id=\"sidebar\">\n"
        xhtml += self.__generate_sidebar_fileinfo()
        xhtml += self.__generate_sidebar_foldertree(entrystore)
        xhtml += self.__generate_sidebar_accountlist(entrystore)
        xhtml += "</div>\n"

        return xhtml


    def __generate_sidebar_accountlist(self, entrystore):
        "Generates an account list"

        # find the entries
        entries = {}
        iter = entrystore.iter_nth_child(None, 0)

        while iter is not None:
            e = entrystore.get_entry(iter)
            e.path = self.__get_entryid(entrystore, iter)

            if type(e) != entry.FolderEntry:
                if type(e) not in entries:
                    entries[type(e)] = []

                entries[type(e)].append(e)

            iter = entrystore.iter_traverse_next(iter)


        # generate the xhtml
        xhtml = ""
        xhtml += (" <h2><img src=\"%s/sidebar/accountlist.png\" class=\"icon\" alt=\"Account list\" />" + _('Account list') + "</h2>\n") % IMAGEPATH
        xhtml += "\n"

        for entrytype in entry.ENTRYLIST:
            if entrytype not in entries:
                continue

            xhtml += "  <h3>%s:</h3>\n" % entrytype().typename
            xhtml += "\n"
            xhtml += "  <ul class=\"accountlist\">\n"

            entrylist = entries[entrytype]

            for e in entrylist:
                xhtml += "      <li><a href=\"#%s\">%s</a></li>\n" % ( str(e.path), e.name )

            xhtml += "  </ul>\n"
            xhtml += "\n"

        return xhtml


    def __generate_sidebar_fileinfo(self):
        "Generates file info for the sidebar"

        xhtml = ""
        xhtml += "  <h2 style=\"margin-top: 0px;\"><img src=\"%s/sidebar/file.png\" class=\"icon\" alt=\"File info\" />File info</h2>\n" % IMAGEPATH
        xhtml += "\n"
        xhtml += "  <h3>" + _('Created:') + "</h3>\n"
        xhtml += "  <p>%s</p>\n" % time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
        xhtml += "\n"
        xhtml += "  <h3>" + _('Revelation version:') + "</h3>\n"
        xhtml += "  <p>%s</p>\n" % config.VERSION
        xhtml += "\n"

        return xhtml


    def __generate_sidebar_foldertree(self, entrystore, parent = None, depth = 0):
        "Generates a folder tree for the sidebar"

        tabs = "\t" * ((depth * 2) - 2)
        xhtml = ""

        if parent is None:
            xhtml += "  <h2><img src=\"%s/sidebar/foldertree.png\" class=\"icon\" alt=\"Folder tree\" />Folder tree</h2>\n" % IMAGEPATH
            xhtml += "\n"

        # fetch folder list
        folders = []
        for i in range(entrystore.iter_n_children(parent)):
            iter = entrystore.iter_nth_child(parent, i)
            e = entrystore.get_entry(iter)
            e.path = self.__get_entryid(entrystore, iter)
            e.iter = iter

            if type(e) == entry.FolderEntry:
                folders.append(e)


        # generate xhtml
        if len(folders) > 0:
            xhtml += tabs + "<ul>\n"

            for e in folders:
                childxhtml = self.__generate_sidebar_foldertree(entrystore, e.iter, depth + 1)

                if childxhtml != "":
                    xhtml += tabs + "   <li>\n"
                    xhtml += tabs + "       <a href=\"#%s\">%s</a>\n" % ( str(e.path), e.name )
                    xhtml += childxhtml
                    xhtml += tabs + "   </li>\n"

                else:
                    xhtml += tabs + "   <li><a href=\"#%s\">%s</a></li>\n" % ( str(e.path), e.name )

            xhtml += tabs + "</ul>\n"


        return xhtml


    def __get_entryid(self, entrystore, iter):
        "Returns an entry id for an iter"

        path = "entry-"

        for i in entrystore.get_path(iter):
            path += str(i) + "-"

        path = path[:-1]

        return path


    def export_data(self, entrystore, password = None):
        "Exports data from an entrystore into a XHTML document"

        xhtml = ""
        xhtml += self.__generate_header()
        xhtml += self.__generate_sidebar(entrystore)
        xhtml += self.__generate_entry(entrystore, None)
        xhtml += self.__generate_footer()

        return xhtml




#
# Revelation - a password manager for GNOME 2
# http://oss.codepoet.no/revelation/
# $Id$
#
# Module for data handlers
#
#
# Copyright (c) 2003-2006 Erik Grinaker
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from .base import Error, DataError, FormatError, PasswordError, VersionError

from .csvfile import CSV, Bitwarden
from .fpm import FPM
from .gpass import GPass04, GPass05
from .keepassxc import KeepassXCCSV
from .netrc import NetRC
from .pwsafe import PasswordSafe1, PasswordSafe2, MyPasswordSafe, MyPasswordSafeOld, PasswordGorilla
from .rvl import RevelationXML, Revelation, Revelation2, RevelationLUKS
from .splashid import SplashIDCSV
from .text import PlainText
from .xhtml import XHTML

HANDLERS = [
    CSV,
    Bitwarden,
    FPM,
    GPass04,
    GPass05,
    KeepassXCCSV,
    MyPasswordSafe,
    MyPasswordSafeOld,
    NetRC,
    PasswordGorilla,
    PasswordSafe1,
    PasswordSafe2,
    PlainText,
    Revelation,
    Revelation2,
    RevelationLUKS,
    SplashIDCSV,
    XHTML,
    RevelationXML
]


class DetectError(Error):
    "Exception for autodetection error"
    pass


def detect_handler(input):
    "Detects which handler may process a data stream"

    for handler in get_import_handlers():
        if handler().detect(input) == True:
            return handler

    else:
        raise DetectError


def get_export_handlers():
    "Returns a list of handlers which can export"

    handlers = []

    for handler in HANDLERS:
        if handler.exporter:
            handlers.append(handler)

    return handlers


def get_import_handlers():
    "Returns a list of handlers which can import"

    handlers = []

    for handler in HANDLERS:
        if handler.importer:
            handlers.append(handler)

    return handlers




#
# Revelation - a password manager for GNOME 2
# http://oss.codepoet.no/revelation/
# $Id$
#
# Module for IO-related functionality
#
#
# Copyright (c) 2003-2006 Erik Grinaker
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from . import datahandler

import os.path, re
from gi.repository import Gio, GObject, GLib

class DataFile(GObject.GObject):
    "Handles data files"

    def __init__(self, handler):
        GObject.GObject.__init__(self)

        self.__uri      = None
        self.__handler      = None
        self.__password     = None
        self.__monitorhandle    = None

        self.set_handler(handler)


    def __str__(self):
        return self.get_file() or ""


    def __cb_monitor(self, monitor_uri, info_uri, event, data = None):
        "Callback for file monitoring"
        if event == Gio.FileMonitorEvent.CHANGED:
            self.emit("content-changed", self.get_file())


    def __monitor(self, file):
        "Starts monitoring a file"

        self.__monitor_stop()

        if file != None:
            self.__monitorhandle = file_monitor(file, self.__cb_monitor)


    def __monitor_stop(self):
        "Stops monitoring the current file"

        if self.__monitorhandle != None:
            file_monitor_cancel(self.__monitorhandle)
            self.__monitorhandle = None


    def close(self):
        "Closes the current file"

        self.set_password(None)
        self.set_file(None)


    def get_file(self):
        "Gets the current file"

        return self.__uri and re.sub("^file://", "", str(self.__uri)) or None


    def get_handler(self):
        "Gets the current handler"

        return self.__handler


    def get_password(self):
        "Gets the current password"

        return self.__password


    def load(self, file, password = None, pwgetter = None):
        "Loads a file"

        file = file_normpath(file)
        data = file_read(file)

        if self.__handler is None:
            self.__handler = datahandler.detect_handler(data)()

        self.__handler.check(data)

        if self.__handler.encryption == True and password is None and pwgetter != None:
            password = pwgetter()

        entrystore = self.__handler.import_data(data, password)

        self.set_password(password)
        self.set_file(file)

        return entrystore


    def save(self, entrystore, file, password = None):
        "Saves an entrystore to a file"

        self.__monitor_stop()
        file_write(file, self.__handler.export_data(entrystore, password))

        # need to use idle_add() to avoid notifying about current save
        GLib.idle_add(lambda: self.__monitor(file))

        self.set_password(password)
        self.set_file(file)


    def set_file(self, file):
        "Sets the current file"

        uri = file is not None and file_normpath(file) or None

        if self.__uri != uri:
            self.__uri = uri
            self.emit("changed", file)

            self.__monitor(file)


    def set_handler(self, handler):
        "Sets and initializes the current data handler"

        self.__handler = handler is not None and handler() or None


    def set_password(self, password):
        "Sets the password for the current file"

        self.__password = password


GObject.type_register(DataFile)
GObject.signal_new("changed", DataFile, GObject.SignalFlags.ACTION,
                   GObject.TYPE_BOOLEAN, (str,))
GObject.signal_new("content-changed", DataFile, GObject.SignalFlags.ACTION,
                   GObject.TYPE_BOOLEAN, (str,))



def file_exists(file):
    "Checks if a file exists"

    if file is None:
        return False

    return Gio.File.new_for_path(file).query_exists()


def file_is_local(file):
    "Checks if a file is on a local filesystem"

    if file is None:
        return False

    return Gio.File.new_for_path(file).get_uri_scheme() == 'file'


def file_monitor(file, callback):
    "Starts monitoring a file"

    try:
        handle = Gio.File.new_for_path(file).monitor_file(Gio.FileMonitorFlags.NONE, None)
        handle.connect('changed', callback)
        return handle
    except GLib.GError:
        return None


def file_monitor_cancel(handle):
    "Cancels file monitoring"

    handle.cancel()


def file_normpath(file):
    "Normalizes a file path"

    if file in ( None, "" ):
        return ""

    file = re.sub("^file:/{,2}", "", file)
    file = os.path.expanduser(file)

    if not re.match("^[a-zA-Z]+://", file) and file[0] != "/":
        file = os.path.abspath(file)

    # Does URI() do anything useful with the input? -thomas jenkins
    #return re.sub("^file:/{,2}", "", str(gnomevfs.URI(file)))
    return file


def file_read(file):
    "Reads data from a file"

    try:
        if file is None:
            raise IOError

        ok, contents, etag = Gio.File.new_for_path(file).load_contents()
        return contents

    except GLib.GError:
        raise IOError


def file_write(file, data):
    "Writes data to file"

    try:
        if file is None:
            raise IOError

        if data is None:
            data = ""

        if isinstance(data, str):
            data = data.encode()

        ok, etag = Gio.File.new_for_path(file).replace_contents(data, None, True, Gio.FileCreateFlags.REPLACE_DESTINATION, None)
        return ok

    except GLib.GError:
        raise IOError

